"""
Helpers to analyze the metamodel IR in ways that are specific to Rel generation.
"""
from __future__ import annotations

from typing import cast, Tuple
from relationalai.early_access.metamodel import ir, util, visitor, builtins
from relationalai.early_access.metamodel.util import OrderedSet


def is_external(relation: ir.Relation):
   return builtins.external_annotation in relation.annotations

def effective_logicals(tasks: OrderedSet[ir.Task]) -> OrderedSet[ir.Logical]:
    return util.OrderedSet.from_iterable(filter(lambda t: is_effective_logical(t), tasks))

def is_effective_logical(n: ir.Task):
    return isinstance(n, ir.Logical) and len(visitor.collect_by_type(ir.Update, n)) > 0

def nullable_logicals(tasks: OrderedSet[ir.Task]) -> OrderedSet[ir.Logical]:
    return util.OrderedSet.from_iterable(filter(lambda t: is_nullable_logical(t), tasks))

def is_nullable_logical(n: ir.Task):
    return isinstance(n, ir.Logical) and any(isinstance(v, ir.Default) and v.value is None for v in n.hoisted)

def hoisted_vars(hoisted: Tuple[ir.VarOrDefault, ...]) -> list[ir.Var]:
    return [v.var if isinstance(v, ir.Default) else v for v in hoisted]

def vars(args: Tuple[ir.Value, ...]) -> list[ir.Var]:
    return cast(list[ir.Var], list(filter(lambda v: isinstance(v, ir.Var), args)))

def outer_join_prefix(nullable_logicals: OrderedSet[ir.Logical], groups: dict[str, OrderedSet[ir.Task]]) -> list[ir.Var]:
    """ Check if it is possible to use an outer join on the logical that has these nullable
    logicals and these groups of tasks. This function returns a list of prefix variables
    for the outer join.

    There are several requirements to use the outer join. If some is not met, the
    function returns an empty list
    """
    # all nested logicals must be nullable
    if len(nullable_logicals) != len(groups["logicals"]):
        return []

    # at this point, all nested logicals are nullable, but they should only have a single hoisted variable
    for logical in nullable_logicals:
        if len(logical.hoisted) != 1:
            return []

    # outer joins only work on outputs
    if "output" not in groups or len(groups["output"]) != 1:
        return []
    output = cast(ir.Output, groups["output"].some())

    # the output variables must be a prefix + nullable variables; the length of the
    # prefix is the number of extra aliases in the output
    prefix_length = len(output.aliases) - len(nullable_logicals)
    if prefix_length < 1:
        return []

    # get the hoisted vars from the nullable logicals, a bit messy
    logical_hoisted_vars = OrderedSet.from_iterable([cast(ir.Default, logical.hoisted[0]).var for logical in nullable_logicals])
    prefix=[]
    i = 0
    for _, var in output.aliases:
        if i < prefix_length:
            # the first prefix_length aliases are the prefix variables
            prefix.append(var)
            i += 1
        else:
            # the remaining output variables should be exposed one by one by the logicals
            if var not in logical_hoisted_vars:
                return []
            logical_hoisted_vars.remove(var)
    # basically an assertion as this should be empty if we got here
    if logical_hoisted_vars:
        return []

    # all nullable logicals must join with all the prefix variables
    for logical in nullable_logicals:
        vars = visitor.collect_implicit_vars(logical)
        for prefix_var in prefix:
            if prefix_var not in vars:
                return []

    return prefix



from relationalai.early_access.dsl.core.types.standard import String, Float, Integer, Date, DateTime
from relationalai.early_access.dsl.ontologies.models import Model
from relationalai.early_access.dsl.types.entities import EntityType
from owlready2 import owl, get_ontology
from datetime import datetime, date
import re


class OwlAdapter:

    def __init__(self, owl_file_path: str):
        self.ontology = get_ontology(owl_file_path).load()
        self._prefix = f"{self.ontology.name.lower()}."
        self._reading_pattern = re.compile(r'(?<!^)(?=[A-Z])')
        self.model = self._bootstrap_model()
        self._add_classes()
        self._add_object_properties()
        self._add_datatype_properties()
        self._add_individuals()

    def _clean_name(self, name: str):
        cs = name.lower().removeprefix(self._prefix)
        return cs

    @staticmethod
    def _is_root_class(ontology_class):
        return not ontology_class.is_a or (owl.Thing in ontology_class.is_a)

    @staticmethod
    def _is_functional(object_property):
        return owl.FunctionalProperty in object_property.is_a

    @staticmethod
    def _is_inverse_functional(object_property):
        return owl.InverseFunctionalProperty in object_property.is_a

    def _get_reading(self, relationship_name: str):
        return self._reading_pattern.sub(' ', relationship_name).lower()

    def _get_mandatory_participation(self, c):
        properties = []
        pattern = r'Inverse\((.*?)\)'

        if c is not None:
            for isa in c.is_a:
                if "some" in str(isa):
                    match = re.search(pattern, str(isa.property))
                    if match:
                        result = match.group(1)
                        properties.append(self._clean_name(result + "-"))
                    else:
                        properties.append(self._clean_name(str(isa.property)))
        return properties

    @staticmethod
    def _xsd_type_to_rel_type(tp):
        mapping = {
            str: String,
            int: Integer,
            float: Float,
            date: Date,
            datetime: DateTime
        }
        return mapping.get(tp, String)

    @staticmethod
    def _get_covering_subclasses(c):
        result = []
        for ec in c.equivalent_to:
            if "|" in str(ec):
                for cs in ec.Classes:
                    result.append(cs)
        return result

    def _get_disjoint_classes(self, c):
        disjoints = []
        pattern = re.compile(r'Not\((.*?)\)')

        for isa in c.is_a:
            match = pattern.search(str(isa))
            if match:
                disjoints.append(self._clean_name(match.group(1)))
            else:
                disjoints.append(self._clean_name(str(getattr(isa, 'property', isa))))

        return disjoints

    def _has_disjoint_sub_types(self, c):
        ecu = sorted([self._clean_name(str(ec)) for ec in self._get_covering_subclasses(c)])
        for ec in self._get_covering_subclasses(c):
            ecu_compare = [str(df) for df in self._get_disjoint_classes(ec)]
            ecu_compare.append(self._clean_name(str(ec)))
            if not ecu == sorted(ecu_compare):
                return False
        return True

    def _bootstrap_model(self) -> Model:
        model_name = re.sub(r'(?<!^)(?=[A-Z])', '_', self.ontology.name).lower()
        md = Model(model_name + "Model")

        # IRI ValueType declaration - this will be used as preferred identifier for all entity types
        iri = md.value_type("IRI", String)
        # OWL Things Entity declaration - all the root classes will be subtype of OWL Thing
        # This is needed because in Owl individuals can have more than one type; in such
        # cases we will assign them the generic OWL Thing type.
        md.entity_type("OWLThing", iri)
        return md

    # Classes and Subclasses
    def _add_classes(self):
        owl_thing_entity = self.model.lookup_concept("OWLThing")
        owl_thing_subclasses = []
        for ontology_class in list(self.ontology.classes()):
            class_name = self._clean_name(str(ontology_class))
            class_entity = self.model.lookup_concept(class_name)

            if class_entity is None:
                # Classes that do not have any super type are identified by the IRI value type
                # Re-declaring the identifier here even if it is technically inherited from OWL Thing as this
                # class is just a placeholder, and we might drop it later

                class_entity = self.model.entity_type(class_name)
                if self._is_root_class(ontology_class):
                    owl_thing_subclasses.append(class_entity)

            # SubClasses
            subclasses = []
            for sub_class in list(ontology_class.subclasses()):
                sub_class_name = self._clean_name(str(sub_class))
                sub_class_entity = self.model.lookup_concept(sub_class_name)
                if sub_class_entity is None:
                    sub_class_entity = self.model.entity_type(sub_class_name)
                subclasses.append(sub_class_entity)

            # Check for inclusive or exclusive type arrows
            if subclasses and self._has_disjoint_sub_types(ontology_class) and self._get_covering_subclasses(ontology_class):
                self.model.subtype_arrow(class_entity, subclasses, True)
            elif subclasses and self._get_covering_subclasses(ontology_class):
                self.model.subtype_arrow(class_entity, subclasses, False, True)
            elif subclasses:
                self.model.subtype_arrow(class_entity, subclasses)

        if owl_thing_entity is not None:
            self.model.subtype_arrow(owl_thing_entity, owl_thing_subclasses)

    # Object Properties
    def _add_object_properties(self):
        for op in list(self.ontology.object_properties()):
            object_property_name = self._clean_name(str(op))
            # In OWL, in general domains and ranges can be defined as union or intersection of multiple classes.
            # This has no direct translation to our model, thus we can only cover the cases where the domains and
            # ranges are single expression. We need to check for this here.
            if len(op.domain) == 1 and len(op.range) == 1:
                d = self._clean_name(str(op.domain[0]))
                domain_entity = self.model.lookup_concept(d)
                r = self._clean_name(str(op.range[0]))
                range_entity = self.model.lookup_concept(r)

                inverse_property_name = ""
                if op.inverse_property is not None:
                    inverse_property_name = self._clean_name(str(op.inverse_property))

                with (self.model.relationship() as rel):
                    # Check for functionality and mandatory participation for the domain
                    if object_property_name in self._get_mandatory_participation(
                            self.model.lookup_concept(op.domain[0])) and self._is_functional(op):
                        rel.role(domain_entity, mandatory=True, unique=True)
                    elif self._is_functional(op):
                        rel.role(domain_entity, unique=True)
                    else:
                        rel.role(domain_entity, name=object_property_name)
                    # Check for functionality and mandatory participation for the range
                    if object_property_name + '-' in self._get_mandatory_participation(
                            self.model.lookup_concept(op.range[0])) and self._is_inverse_functional(op):
                        rel.role(range_entity, mandatory=True, unique=True)
                    elif self._is_inverse_functional(op):
                        rel.role(range_entity, unique=True)
                    else:
                        rel.role(range_entity)

                    # Readings
                    rel.relation(rel.role_at(0), self._get_reading(object_property_name), rel.role_at(1))
                    if inverse_property_name != "":
                        rel.relation(rel.role_at(1), self._get_reading(inverse_property_name), rel.role_at(0))

                    if not self._is_functional(op) and not self._is_inverse_functional(op):
                        self.model.unique(rel.role_at(0), rel.role_at(1))
            elif len(op.domain) == 1:
                Exception(
                    f"++ Error: Unsupported complex expression in the domain of object property {object_property_name}")
            else:
                Exception(
                    f"++ Error: Unsupported complex expression in the range of object property {object_property_name}")

    # Datatype Properties
    def _add_datatype_properties(self):
        for dp in list(self.ontology.data_properties()):
            datatype_property_name = self._clean_name(str(dp))
            datatype_property_name = datatype_property_name[:1].upper() + datatype_property_name[1:]
            data_type_entity = self.model.lookup_concept(datatype_property_name)
            if data_type_entity is None and len(dp.range) == 1:
                r = dp.range[0]
                xsd_type = self._xsd_type_to_rel_type(r)
                data_type_entity = self.model.value_type(datatype_property_name, xsd_type)
            else:
                data_type_entity = self.model.value_type(datatype_property_name, String)

            if len(dp.domain) == 1:
                d = self._clean_name(str(dp.domain[0]))
                domain_entity = self.model.lookup_concept(d)

                with self.model.relationship(domain_entity, "has", data_type_entity) as rel:
                    self.model.unique(rel.role_at(0), rel.role_at(1))
            else:
                Exception(
                    f"++ Error: Unsupported complex expression in the domain of datatype {datatype_property_name}")

    # Individuals
    def _add_individuals(self):
        for idv in list(self.ontology.individuals()):
            iri = idv.namespace.base_iri + idv.name
            individual_class = ""
            if len(idv.is_a) == 1:
                individual_class = self._clean_name(str(idv.is_a[0]))
            else:
                individual_class += "OWLThing"
            entity_type = self.model.lookup_concept(individual_class)
            if isinstance(entity_type, EntityType):
                entity_type._add_instance(iri)



"""
    Support for traversing the IR, often to search for information.
"""
from dataclasses import dataclass, field
from typing import Callable, Optional, TypeVar, cast, Tuple, Union as PyUnion, Generic
from abc import abstractmethod
from .util import OrderedSet, ordered_set, FrozenOrderedSet, rewrite_list, rewrite_set
from . import ir

#--------------------------------------------------
# Visitor Abstraction
#--------------------------------------------------

Result = TypeVar('Result')
class GenericVisitor(Generic[Result]):
    """
    Abstract visitor with handlers for each IR node type.
    Each handler should return a value of type `Result`.

    Actual behavior (e.g., AST traversal) should be implemented in
    subclasses of `GenericVisitor`.
    """

    ##################################################
    # Model

    @abstractmethod
    def visit_model(self, node: ir.Model, parent: Optional[ir.Node]) -> Result:
        pass

    ##################################################
    # Engine

    @abstractmethod
    def visit_capability(self, node: ir.Capability, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_engine(self, node: ir.Engine, parent: Optional[ir.Node]) -> Result:
        pass

    ##################################################
    # Types

    @abstractmethod
    def visit_scalartype(self, node: ir.ScalarType, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_listtype(self, node: ir.ListType, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_settype(self, node: ir.SetType, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_uniontype(self, node: ir.UnionType, parent: Optional[ir.Node]) -> Result:
        pass

    ##################################################
    # Relations

    @abstractmethod
    def visit_field(self, node: ir.Field, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_relation(self, node: ir.Relation, parent: Optional[ir.Node]) -> Result:
        pass

    ##################################################
    # Variables and values

    @abstractmethod
    def visit_var(self, node: ir.Var, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_default(self, node: ir.Default, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_literal(self, node: ir.Literal, parent: Optional[ir.Node]) -> Result:
        pass

    ##################################################
    # Composite tasks

    @abstractmethod
    def visit_logical(self, node: ir.Logical, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_union(self, node: ir.Union, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_sequence(self, node: ir.Sequence, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_match(self, node: ir.Match, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_until(self, node: ir.Until, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_wait(self, node: ir.Wait, parent: Optional[ir.Node]) -> Result:
        pass

    ##################################################
    # Logical tasks

    @abstractmethod
    def visit_not(self, node: ir.Not, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_exists(self, node: ir.Exists, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_forall(self, node: ir.ForAll, parent: Optional[ir.Node]) -> Result:
        pass

    ##################################################
    # Iteration tasks

    @abstractmethod
    def visit_loop(self, node: ir.Loop, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_break(self, node: ir.Break, parent: Optional[ir.Node]) -> Result:
        pass

    ##################################################
    # Leaf tasks

    @abstractmethod
    def visit_update(self, node: ir.Update, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_annotation(self, node: ir.Annotation, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_lookup(self, node: ir.Lookup, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_output(self, node: ir.Output, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_construct(self, node: ir.Construct, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_aggregate(self, node: ir.Aggregate, parent: Optional[ir.Node]) -> Result:
        pass

    @abstractmethod
    def visit_task(self, node: ir.Task, parent: Optional[ir.Node]) -> Result:
        pass

@dataclass
class Visitor(GenericVisitor[None]):
    """
    A visitor that just walks the AST, performing no changes.
    """

    def enter(self, node: ir.Node, parent: Optional[ir.Node]=None) -> "Visitor":
        """ Visit the node, possibly returning a new visitor to be used to visit the node's children. """
        return self

    def prune(self):
        """ Return a new Visitor that just returns each node unchanged without traversing its children. """
        return PruneVisitor()

    def leave(self, node: ir.Node, parent: Optional[ir.Node]=None) -> ir.Node:
        """ Visit the node after visiting it and its children. """
        return node

    def _walk_node(self, node: ir.Node, parent: Optional[ir.Node]=None):
        # Enter the node, returning a new context.
        v = self.enter(node, parent)

        # Dispatch to the handler for the node.
        # By default the handler traverses the node's children and reconstructs the node.
        node.accept(v, parent)

        self.leave(node, parent)

    # Avoid dispatch overhead for some types.
    # These should never be called directly.
    def _walk_var(self, node: ir.Var, parent: Optional[ir.Node]):
        v = self.enter(node, parent)
        v.visit_var(node, parent)
        self.leave(node, parent)

    def _walk_var_or_default(self, node: PyUnion[ir.Var, ir.Default], parent: Optional[ir.Node]):
        v = self.enter(node, parent)
        if isinstance(node, ir.Var):
            v.visit_var(node, parent)
        else:
            v.visit_default(node, parent)
        self.leave(node, parent)

    def _walk_value(self, node: ir.Value, parent: Optional[ir.Node]):
        if node is None:
            return None
        elif isinstance(node, ir.Var):
            return self._walk_var(node, parent)
        elif isinstance(node, ir.Relation):
            return self._walk_relation(node, parent)
        elif isinstance(node, ir.Node):
            return self._walk_node(node, parent)
        elif isinstance(node, Tuple):
            for item in node:
                self._walk_value(item, parent)
        elif isinstance(node, FrozenOrderedSet):
            for item in node:
                self._walk_value(item, parent)

    def _walk_engine(self, node: ir.Engine, parent: Optional[ir.Node]):
        v = self.enter(node, parent)
        v.visit_engine(node, parent)
        self.leave(node, parent)

    def _walk_relation(self, node: ir.Relation, parent: Optional[ir.Node]):
        v = self.enter(node, parent)
        v.visit_relation(node, parent)
        self.leave(node, parent)

    def _walk_type(self, node: ir.Type, parent: Optional[ir.Node]):
        # The if-isinstance chain saves about 10% on a traversal vs. node.accept(v, parent)
        v = self.enter(node, parent)
        if isinstance(node, ir.ScalarType):
            v.visit_scalartype(node, parent)
        elif isinstance(node, ir.ListType):
            v.visit_listtype(node, parent)
        elif isinstance(node, ir.SetType):
            v.visit_settype(node, parent)
        elif isinstance(node, ir.UnionType):
            v.visit_uniontype(node, parent)
        else:
            raise NotImplementedError(f"visit_type not implemented for {type(node)}")
        self.leave(node, parent)

    ##################################################
    # Model

    def visit_model(self, node: ir.Model, parent: Optional[ir.Node]):
        for c in node.engines:
            self._walk_engine(c, node)
        for c in node.relations:
            self._walk_relation(c, node)
        for c in node.types:
            self._walk_type(c, node)
        self._walk_node(node.root, node)

    ##################################################
    # Engine

    def visit_capability(self, node: ir.Capability, parent: Optional[ir.Node]):
        pass

    def visit_engine(self, node: ir.Engine, parent: Optional[ir.Node]):
        for c in node.capabilities:
            self._walk_node(c, node)
        for c in node.relations:
            self._walk_relation(c, node)

    ##################################################
    # Types

    def visit_scalartype(self, node: ir.ScalarType, parent: Optional[ir.Node]):
        for t in node.super_types:
            self._walk_type(t, node)

    def visit_listtype(self, node: ir.ListType, parent: Optional[ir.Node]):
        self._walk_type(node.element_type, node)

    def visit_settype(self, node: ir.SetType, parent: Optional[ir.Node]):
        self._walk_type(node.element_type, node)

    def visit_uniontype(self, node: ir.UnionType, parent: Optional[ir.Node]):
        for t in node.types:
            self._walk_type(t, node)

    ##################################################
    # Relations

    def visit_field(self, node: ir.Field, parent: Optional[ir.Node]):
        self._walk_type(node.type, node)

    def visit_relation(self, node: ir.Relation, parent: Optional[ir.Node]):
        for f in node.fields:
            self._walk_node(f, node)
        for r in node.requires:
            self._walk_node(r, node)

    ##################################################
    # Variables and values

    def visit_var(self, node: ir.Var, parent: Optional[ir.Node]):
        self._walk_type(node.type, node)

    def visit_default(self, node: ir.Default, parent: Optional[ir.Node]):
        self._walk_var(node.var, node)
        self._walk_value(node.value, node)

    def visit_literal(self, node: ir.Literal, parent: Optional[ir.Node]):
        self._walk_type(node.type, node)

    ##################################################
    # Composite tasks

    def visit_logical(self, node: ir.Logical, parent: Optional[ir.Node]):
        if node.engine is not None:
            self._walk_engine(node.engine, node)
        for h in node.hoisted:
            self._walk_var_or_default(h, node)
        for b in node.body:
            self._walk_node(b, node)

    def visit_union(self, node: ir.Union, parent: Optional[ir.Node]):
        if node.engine is not None:
            self._walk_engine(node.engine, node)
        for h in node.hoisted:
            self._walk_var_or_default(h, node)
        for t in node.tasks:
            self._walk_node(t, node)

    def visit_sequence(self, node: ir.Sequence, parent: Optional[ir.Node]):
        if node.engine is not None:
            self._walk_engine(node.engine, node)
        for h in node.hoisted:
            self._walk_var_or_default(h, node)
        for t in node.tasks:
            self._walk_node(t, node)

    def visit_match(self, node: ir.Match, parent: Optional[ir.Node]):
        if node.engine is not None:
            self._walk_engine(node.engine, node)
        for h in node.hoisted:
            self._walk_var_or_default(h, node)
        for t in node.tasks:
            self._walk_node(t, node)

    def visit_until(self, node: ir.Until, parent: Optional[ir.Node]):
        if node.engine is not None:
            self._walk_engine(node.engine, node)
        for h in node.hoisted:
            self._walk_var_or_default(h, node)
        self._walk_node(node.check, node)
        self._walk_node(node.body, node)

    def visit_wait(self, node: ir.Wait, parent: Optional[ir.Node]):
        if node.engine is not None:
            self._walk_engine(node.engine, node)
        for h in node.hoisted:
            self._walk_var_or_default(h, node)
        self._walk_node(node.check, node)

    ##################################################
    # Logical tasks

    def visit_not(self, node: ir.Not, parent: Optional[ir.Node]):
        if node.engine is not None:
            self._walk_engine(node.engine, node)
        self._walk_node(node.task, node)

    def visit_exists(self, node: ir.Exists, parent: Optional[ir.Node]):
        if node.engine is not None:
            self._walk_engine(node.engine, node)
        for v in node.vars:
            self._walk_var(v, node)
        self._walk_node(node.task, node)

    def visit_forall(self, node: ir.ForAll, parent: Optional[ir.Node]):
        if node.engine is not None:
            self._walk_engine(node.engine, node)
        for v in node.vars:
            self._walk_var(v, node)
        self._walk_node(node.task, node)

    ##################################################
    # Iteration tasks

    def visit_loop(self, node: ir.Loop, parent: Optional[ir.Node]):
        if node.engine is not None:
            self._walk_engine(node.engine, node)
        for h in node.hoisted:
            self._walk_var_or_default(h, node)
        self._walk_var(node.iter, node)
        self._walk_node(node.body, node)

    def visit_break(self, node: ir.Break, parent: Optional[ir.Node]):
        if node.engine is not None:
            self._walk_engine(node.engine, node)
        self._walk_node(node.check, node)

    ##################################################
    # Leaf tasks

    def visit_update(self, node: ir.Update, parent: Optional[ir.Node]):
        if node.engine is not None:
            self._walk_engine(node.engine, node)
        self._walk_relation(node.relation, node)
        for a in node.args:
            self._walk_value(a, node)
        for a in node.annotations:
            self._walk_node(a, node)

    def visit_annotation(self, node: ir.Annotation, parent: Optional[ir.Node]):
        self._walk_relation(node.relation, node)
        for a in node.args:
            self._walk_value(a, node)

    def visit_lookup(self, node: ir.Lookup, parent: Optional[ir.Node]):
        if node.engine is not None:
            self._walk_engine(node.engine, node)
        self._walk_relation(node.relation, node)
        for a in node.args:
            self._walk_value(a, node)

    def visit_output(self, node: ir.Output, parent: Optional[ir.Node]):
        if node.engine is not None:
            self._walk_engine(node.engine, node)
        for _, v in node.aliases:
            self._walk_var(v, node)

    def visit_construct(self, node: ir.Construct, parent: Optional[ir.Node]):
        if node.engine is not None:
            self._walk_engine(node.engine, node)
        for v in node.values:
            self._walk_value(v, node)
        self._walk_var(node.id_var, node)

    def visit_aggregate(self, node: ir.Aggregate, parent: Optional[ir.Node]):
        if node.engine is not None:
            self._walk_engine(node.engine, node)
        self._walk_relation(node.aggregation, node)
        for p in node.projection:
            self._walk_var(p, node)
        for g in node.group:
            self._walk_var(g, node)
        for a in node.args:
            self._walk_value(a, node)

    def visit_task(self, node: ir.Task, parent: Optional[ir.Node]):
        if node.engine is not None:
            self._walk_engine(node.engine, node)

#--------------------------------------------------
# Some generally useful visitors
#--------------------------------------------------

@dataclass
class PruneVisitor(Visitor):
    """ A Visitor that does not walk children. """

    def _walk_node(self, node: ir.Node, parent: Optional[ir.Node]=None):
        pass
    def _walk_var(self, node: ir.Var, parent: Optional[ir.Node]):
        pass
    def _walk_var_or_default(self, node: ir.VarOrDefault, parent: Optional[ir.Node]):
        pass
    def _walk_engine(self, node: ir.Engine, parent: Optional[ir.Node]):
        pass
    def _walk_relation(self, node: ir.Relation, parent: Optional[ir.Node]):
        pass
    def _walk_type(self, node: ir.Type, parent: Optional[ir.Node]):
        pass

@dataclass
class DAGVisitor(Visitor):
    """ A visitor that just walks the AST, performing no changes. Nodes are visited at most once. """
    seen: set[ir.Node] = field(default_factory=set)

    def _walk_node(self, node: ir.Node, parent: Optional[ir.Node]=None):
        if node in self.seen:
            return
        self.seen.add(node)
        return super()._walk_node(node, parent)
    def _walk_var(self, node: ir.Var, parent: Optional[ir.Node]):
        if node in self.seen:
            return
        self.seen.add(node)
        return super()._walk_var(node, parent)
    def _walk_var_or_default(self, node: ir.VarOrDefault, parent: Optional[ir.Node]):
        if node in self.seen:
            return
        self.seen.add(node)
        return super()._walk_var_or_default(node, parent)
    def _walk_engine(self, node: ir.Engine, parent: Optional[ir.Node]):
        if node in self.seen:
            return
        self.seen.add(node)
        return super()._walk_engine(node, parent)
    def _walk_relation(self, node: ir.Relation, parent: Optional[ir.Node]):
        if node in self.seen:
            return
        self.seen.add(node)
        return super()._walk_relation(node, parent)
    def _walk_type(self, node: ir.Type, parent: Optional[ir.Node]):
        if node in self.seen:
            return
        self.seen.add(node)
        return super()._walk_type(node, parent)

class Collector(Visitor):
    """ A visitor that collects instances that match a predicate. """
    def __init__(self, predicate: Callable[[ir.Node, Optional[ir.Node]], bool]):
        super().__init__()
        self.elements: OrderedSet = ordered_set()
        self.predicate = predicate

    def enter(self, node: ir.Node, parent: Optional[ir.Node]=None):
        if self.predicate(node, parent):
            self.elements.add(node)
        return self

def collect(predicate: Callable[[ir.Node, Optional[ir.Node]], bool], *nodes: ir.Node) -> OrderedSet[ir.Node]:
    """ Collect children of node that match the predicate. """
    c = Collector(predicate)
    for n in nodes:
        n.accept(c)
    return c.elements

T = TypeVar('T')
def collect_by_type(t: PyUnion[type[T], tuple[type[T], ...]], *nodes: ir.Node) -> OrderedSet[T]:
    """ Collect instances of the type t by traversing this node and its children. """
    return cast(OrderedSet[T],
        collect(lambda n, parent: isinstance(n, t), *nodes)
    )

def collect_vars(*nodes: ir.Node) -> OrderedSet[ir.Var]:
    """ Collect all Vars starting at this node. """
    return cast(OrderedSet[ir.Var],
        collect_by_type(ir.Var, *nodes)
    )

def collect_quantified_vars(*nodes: ir.Node) -> OrderedSet[ir.Var]:
    """ Collect all Vars that are children of Exists and ForAll. """
    return cast(OrderedSet[ir.Var],
        collect(
            lambda n, parent: isinstance(n, ir.Var) and isinstance(parent, (ir.Exists, ir.ForAll)),
            *nodes)
    )

def collect_aggregate_vars(*nodes: ir.Node) -> OrderedSet[ir.Var]:
    """ Collect vars that are declared by aggregates in Rel (projection + over). """
    return cast(OrderedSet[ir.Var],
        # TODO - when dealing with multiple aggregations we will need to consider groupbys
        collect(
            lambda n, parent:
                # parent of var is an aggregate and either
                isinstance(parent, ir.Aggregate) and isinstance(n, ir.Var) and (
                # var is in the projection
                n in parent.projection or
                # var is an input to the aggregation's relation
                is_aggregate_input(n, parent)
                ),
            *nodes)
    )

def is_aggregate_input(var: ir.Var, agg: ir.Aggregate):
    # field at the var index is input (the +2 is for projection + group)
    return (var in agg.args and agg.aggregation.fields[agg.args.index(var) + 2].input)

def collect_implicit_vars(*nodes: ir.Node) -> OrderedSet[ir.Var]:
    """ Collect vars except the quantified and aggregate vars. """
    if not nodes:
        return ordered_set()
    return collect_vars(*nodes) - collect_quantified_vars(*nodes) - collect_aggregate_vars(*nodes)



@dataclass
class ReadWriteVisitor(Visitor):
    """
    Compute the set of reads and writes for Logical nodes.

    Note that reads are Lookups and writes are Updates. We don't consider Output a write
    because it is not targeting a relation.
    """
    # TODO: we currently only compute for Logical nodes, but it may be useful for other nodes
    _reads: dict[ir.Logical, OrderedSet[ir.Relation]] = field(default_factory=dict)
    _writes: dict[ir.Logical, OrderedSet[ir.Relation]] = field(default_factory=dict)

    def reads(self, key: ir.Logical):
        # TODO - use a singleton empty set
        return self._reads[key] if key in self._reads else ordered_set()

    def writes(self, key: ir.Logical):
        return self._writes[key] if key in self._writes else ordered_set()

    _stack: list[ir.Logical] = field(default_factory=list)

    def visit_logical(self, node: ir.Logical, parent: Optional[ir.Node]):
        self._stack.append(node)
        super().visit_logical(node, parent)
        self._stack.pop()

    def visit_lookup(self, node: ir.Lookup, parent: Optional[ir.Node]):
        for lu in self._stack:
            if lu not in self._reads:
                self._reads[lu] = ordered_set()
            self._reads[lu].add(node.relation)
        return super().visit_lookup(node, parent)

    def visit_aggregate(self, node: ir.Aggregate, parent: Optional[ir.Node]):
        for lu in self._stack:
            if lu not in self._reads:
                self._reads[lu] = ordered_set()
            self._reads[lu].add(node.aggregation)
        return super().visit_aggregate(node, parent)

    def visit_update(self, node: ir.Update, parent: Optional[ir.Node]):
        for lu in self._stack:
            if lu not in self._writes:
                self._writes[lu] = ordered_set()
            self._writes[lu].add(node.relation)
        return super().visit_update(node, parent)

@dataclass
class DeprecatedPass():
    """
    This is the old Pass class. To be replaced by a Rewriter.

    """

    rewritten: dict[ir.Node, ir.Node] = field(default_factory=dict, init=False, compare=False, hash=False)
    handler_cache: dict[str, Callable[[ir.Node, Optional[ir.Node]], ir.Node]] = field(default_factory=dict, init=False, compare=False, hash=False)

    T = TypeVar('T', bound=PyUnion[ir.Node, ir.Value])
    def walk(self, node: T, parent=None) -> T:
        # if node is a value, just return it
        if not isinstance(node, ir.Node):
            return node

        result = self.rewritten.get(node, None)
        if result:
            return cast(T, result) # type: ignore[reportReturnType]

        # node is actually some Node type, handle with the appropriate handler
        handler = self.handler_cache.get(node.kind, None)
        if not handler:
            handler = getattr(self, f"handle_{node.kind}", None)
            if handler:
                self.handler_cache[node.kind] = handler
        if handler:
            result = cast(ir.Node, handler(node, parent))
            self.rewritten[node] = result
            return cast(T, result) # type: ignore[reportReturnType]
        else:
            raise NotImplementedError(f"walk: {node.kind}")

    def walk_set(self, items: FrozenOrderedSet[T], parent=None) -> FrozenOrderedSet[T]:
        return ordered_set(*[self.walk(n, parent) for n in items]).frozen()

    def walk_list(self, items: Tuple[T, ...], parent=None) -> Tuple[T, ...]:
        return tuple([self.walk(n, parent) for n in items])

    U = TypeVar('U', bound=PyUnion[ir.Node, ir.Value, FrozenOrderedSet[ir.Relation]])

    def walk_dict(self, items: dict[str, U], parent=None) -> dict[str, U]:
        d = dict()
        for k, v in items.items():
            if isinstance(v, FrozenOrderedSet):
                d[k] = self.walk_set(v, parent) # type: ignore
            else:
                d[k] = self.walk(v, parent)
        return d



    #-------------------------------------------------
    # Public Types - Model
    #-------------------------------------------------

    def handle_model(self, model: ir.Model, parent: None):
        engines = rewrite_set(ir.Engine, lambda n: self.walk(n, model), model.engines)
        relations = rewrite_set(ir.Relation, lambda n: self.walk(n, model), model.relations)
        types = rewrite_set(ir.Type, lambda n: self.walk(n, model), model.types)
        root = self.walk(model.root, model)

        return model.reconstruct(engines, relations, types, root)

    #-------------------------------------------------
    # Public Types - Engine
    #-------------------------------------------------

    def handle_capability(self, node: ir.Capability, parent: ir.Node):
        return node

    def handle_engine(self, node: ir.Engine, parent: ir.Node):
        return node

    #-------------------------------------------------
    # Public Types - Data Model
    #-------------------------------------------------

    def handle_scalartype(self, node: ir.ScalarType, parent: ir.Node):
        return node

    def handle_listtype(self, node: ir.ListType, parent: ir.Node):
        return node

    def handle_settype(self, node: ir.SetType, parent: ir.Node):
        return node

    def handle_uniontype(self, node: ir.UnionType, parent: ir.Node):
        # TODO - we could traverse the children
        return node

    def handle_field(self, node: ir.Field, parent: ir.Node):
        return node

    def handle_relation(self, node: ir.Relation, parent: ir.Node):
        return node

    #-------------------------------------------------
    # Public Types - Tasks
    #-------------------------------------------------

    def handle_task(self, node: ir.Task, parent: ir.Node):
        return node

    #
    # Task composition
    #

    def handle_logical(self, node: ir.Logical, parent: ir.Node):
        hoisted = rewrite_list(ir.VarOrDefault, lambda n: self.walk(n, node), node.hoisted)
        body = rewrite_list(ir.Task, lambda n: self.walk(n, node), node.body)

        return node.reconstruct(node.engine, hoisted, body)

    def handle_union(self, node: ir.Union, parent: ir.Node):
        hoisted = rewrite_list(ir.VarOrDefault, lambda n: self.walk(n, node), node.hoisted)
        tasks = rewrite_list(ir.Task, lambda n: self.walk(n, node), node.tasks)

        return node.reconstruct(node.engine, hoisted, tasks)

    def handle_sequence(self, node: ir.Sequence, parent: ir.Node):
        hoisted = rewrite_list(ir.VarOrDefault, lambda n: self.walk(n, node), node.hoisted)
        tasks = rewrite_list(ir.Task, lambda n: self.walk(n, node), node.tasks)

        return node.reconstruct(node.engine, hoisted, tasks)

    def handle_match(self, node: ir.Match, parent: ir.Node):
        hoisted = rewrite_list(ir.VarOrDefault, lambda n: self.walk(n, node), node.hoisted)
        tasks = rewrite_list(ir.Task, lambda n: self.walk(n, node), node.tasks)

        return node.reconstruct(node.engine, hoisted, tasks)

    def handle_until(self, node: ir.Until, parent: ir.Node):
        hoisted = rewrite_list(ir.VarOrDefault, lambda n: self.walk(n, node), node.hoisted)
        check = self.walk(node.check, node)
        body = self.walk(node.body, node)

        return node.reconstruct(node.engine, hoisted, check, body)

    def handle_wait(self, node: ir.Wait, parent: ir.Node):
        hoisted = rewrite_list(ir.VarOrDefault, lambda n: self.walk(n, node), node.hoisted)
        check = self.walk(node.check, node)

        return node.reconstruct(node.engine, hoisted, check)

    #
    # Relational Operations
    #

    def handle_var(self, node: ir.Var, parent: ir.Node):
        type_val = self.walk(node.type, node)
        name_val = node.name or f"v{node.id}"
        return node.reconstruct(type_val, name_val)

    def handle_default(self, node: ir.Default, parent: ir.Node):
        var = self.walk(node.var, node)
        return node.reconstruct(var, node.value)

    def handle_literal(self, node: ir.Literal, parent: ir.Node):
        type_val = self.walk(node.type, node)
        return node.reconstruct(type_val, node.value)

    def handle_annotation(self, node: ir.Annotation, parent: ir.Node):
        return node

    def handle_update(self, node: ir.Update, parent: ir.Node):
        relation = self.walk(node.relation, node)
        args = rewrite_list(ir.Value, lambda n: self.walk(n, node), node.args)
        return node.reconstruct(node.engine, relation, args, node.effect, node.annotations)

    def handle_lookup(self, node: ir.Lookup, parent: ir.Node):
        relation = self.walk(node.relation, node)
        args = rewrite_list(ir.Value, lambda n: self.walk(n, node), node.args)
        return node.reconstruct(node.engine, relation, args)

    def handle_output(self, node: ir.Output, parent: ir.Node):
        def rewrite_alias(pair: Tuple[str, ir.Var]):
            name, x = pair
            new_x = self.walk(x, node)
            if new_x is not x:
                return (name, new_x)
            else:
                return pair
        # rewrite_set can't be passed a generic type, so just use Tuple and cast the result.
        aliases = cast(
            FrozenOrderedSet[Tuple[str, ir.Var]],
            rewrite_set(
                Tuple,
                rewrite_alias,
                node.aliases))
        return node.reconstruct(node.engine, aliases)

    def handle_construct(self, node: ir.Construct, parent: ir.Node):
        values = rewrite_list(ir.Value, lambda n: self.walk(n, node), node.values)
        id_var = self.walk(node.id_var, node)
        return node.reconstruct(node.engine, values, id_var)

    def handle_aggregate(self, node: ir.Aggregate, parent: ir.Node):
        aggregation = self.walk(node.aggregation, node)
        projection = rewrite_list(ir.Var, lambda n: self.walk(n, node), node.projection)
        group = rewrite_list(ir.Var, lambda n: self.walk(n, node), node.group)
        args = rewrite_list(ir.Value, lambda n: self.walk(n, node), node.args)
        return node.reconstruct(node.engine, aggregation, projection, group, args)

    #
    # Logical Quantifiers
    #

    def handle_not(self, node: ir.Not, parent: ir.Node):
        task = self.walk(node.task, node)
        return node.reconstruct(node.engine, task)

    def handle_exists(self, node: ir.Exists, parent: ir.Node):
        vars = rewrite_list(ir.Var, lambda n: self.walk(n, node), node.vars)
        task = self.walk(node.task, node)
        return node.reconstruct(node.engine, vars, task)

    def handle_forall(self, node: ir.ForAll, parent: ir.Node):
        vars = rewrite_list(ir.Var, lambda n: self.walk(n, node), node.vars)
        task = self.walk(node.task, node)
        return node.reconstruct(node.engine, vars, task)

    #
    # Iteration (Loops)
    #
    def handle_loop(self, node: ir.Loop, parent: ir.Node):
        hoisted = rewrite_list(ir.VarOrDefault, lambda n: self.walk(n, node), node.hoisted)
        iter_val = self.walk(node.iter, node)
        body = self.walk(node.body, node)
        return node.reconstruct(node.engine, hoisted, iter_val, body)

    def handle_break(self, node: ir.Break, parent: ir.Node):
        check = self.walk(node.check, node)
        return node.reconstruct(node.engine, check)



#--------------------------------------------------
# FLAGS
#--------------------------------------------------

USE_GRAPH_INDEX = True
WAIT_FOR_STREAM_SYNC = True
USE_PACKAGE_MANAGER = True

#--------------------------------------------------
# Constants
#--------------------------------------------------

RAI_APP_NAME = "RELATIONALAI"
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

class GlobalProfileSetting:
    def __init__(self):
        self.profile = None

    def set(self, profile):
        self.profile = profile

    def get(self):
        return self.profile

GlobalProfile = GlobalProfileSetting()

SNOWFLAKE = "Snowflake"
AZURE = "Azure (Beta)"

SNOWFLAKE_AUTHENTICATOR = {
    "USERNAME & PASSWORD": "snowflake",
    "USERNAME & PASSWORD (MFA ENABLED)": "username_password_mfa",
    "SINGLE SIGN-ON (SSO)": "externalbrowser",
}

FIELD_PLACEHOLDER = ""

SNOWFLAKE_PROFILE_DEFAULTS = {
    "platform": { "value": "snowflake", "required": True },
    "user": { "value": FIELD_PLACEHOLDER, "required": True },
    "account": { "value": FIELD_PLACEHOLDER, "required": True },
    "role": { "value": "PUBLIC", "required": True },
    "warehouse": { "value": FIELD_PLACEHOLDER, "required": True },
    "rai_app_name": { "value": RAI_APP_NAME, "required": True },
    "engine": { "value": FIELD_PLACEHOLDER, "required": False },
    "engine_size": { "value": "HIGHMEM_X64_S", "required": False },
}

SNOWFLAKE_AUTHS = {
    "snowflake": {
        "authenticator": { "value": "snowflake", "required": True },
        "password": { "value": FIELD_PLACEHOLDER, "required": True },
        **SNOWFLAKE_PROFILE_DEFAULTS,
    },
    "username_password_mfa": {
        "authenticator": { "value": "username_password_mfa", "required": True },
        "password": { "value": FIELD_PLACEHOLDER, "required": True },
        "passcode": { "value": FIELD_PLACEHOLDER, "required": False },
        **SNOWFLAKE_PROFILE_DEFAULTS,
    },
    "externalbrowser": {
        "authenticator": { "value": "externalbrowser", "required": True },
        **SNOWFLAKE_PROFILE_DEFAULTS,
    },
}

AZURE_ENVS = {
    "Production": {
        "host": "azure.relationalai.com",
        "client_credentials_url": "https://login.relationalai.com/oauth/token"
    },
    "Early Access": {
        "host": "azure-ea.relationalai.com",
        "client_credentials_url": "https://login-ea.relationalai.com/oauth/token"
    },
    "Staging": {
        "host": "azure-staging.relationalai.com",
        "client_credentials_url": "https://login-staging.relationalai.com/oauth/token"
    },
    "Latest": {
        "host": "azure-latest.relationalai.com",
        "client_credentials_url": "https://login-latest.relationalai.com/oauth/token"
    },
}




from __future__ import annotations
from relationalai.early_access.metamodel import types, factory as f

# Rel Annotations as IR Relations (to be used in IR Annotations)
arrow = f.relation("no_inline", [])
no_diagnostics = f.relation("no_diagnostics", [f.field("code", types.Symbol)])
no_inline = f.relation("no_inline", [])
# inner_loop_non_stratified

annotations = [
    arrow,
    no_diagnostics,
    no_inline,
]



from __future__ import annotations
from dataclasses import dataclass, field
from typing import cast, Optional

from relationalai.early_access.metamodel import ir, factory as f, visitor
from relationalai.early_access.metamodel.compiler import Pass, group_tasks
from relationalai.early_access.metamodel.util import OrderedSet, ordered_set
from relationalai.early_access.rel import metamodel_utils


class Flatten(Pass):
    """
    Traverses the model's root to flatten it as much as possible. The result of this pass is
    a Logical root where all nested tasks that represent a rule in Rel are extraced to the
    top level.

    - nested logical with updates becomes a top-level logical (a rule)

    From:
        Logical
            Logical
                lookup1   <- scope is spread
                Logical
                    lookup2
                    derive foo
                Logical
                    lookup3
                    derive bar
    To:
        Logical
            Logical
                lookup1
                lookup2
                derive foo
            Logical
                lookup1
                lookup3
                derive bar

    - nested logical with aggregates becomes a top-level logical (a rule representing an aggregation)

    From:
        Logical
            Logical
                lookup1
                Logical
                    lookup2
                    aggregate1
                Logical
                    lookup3
                    aggregate2
                output
    To:
        Logical
            Logical
                lookup1
                lookup2
                aggregate1
                derive tmp1
            Logical
                lookup1
                lookup3
                aggregate2
                derive tmp2
            Logical
                lookup1
                lookup tmp1
                lookup tmp2
                output

    - a union becomes a top-level logical for each branch, writing into a temporary relation,
    and a lookup from that relation.

    From:
        Logical
            Logical
                Union
                    Logical
                        lookup1
                    Logical
                        lookup2
                output
    To:
        Logical
            Logical
                lookup1
                derive tmp1
            Logical
                lookup2
                derive tmp1
            Logical
                lookup tmp1
                output

    - a match becomes a top-level logical for each branch, each writing into its own temporary
    relation and a lookup from the last relation. The top-level logical for a branch derives
    into the temporary relation negating the previous branch:

    From:
        Logical
            Logical
                Match
                    Logical
                        lookup1
                    Logical
                        lookup2
                output
    To:
        Logical
            Logical
                lookup1
                derive tmp1
            Logical
                Union            <- tmp1() or (not temp1() and lookup2())
                    lookup tmp1
                    Logical
                        Not
                            lookup tmp1
                        lookup2
                        derive tmp2
            Logical
                lookup tmp2
                output
    """

    #--------------------------------------------------
    # Public API
    #--------------------------------------------------
    def rewrite(self, model: ir.Model, cache) -> ir.Model:

        ctx = Flatten.Context()

        # rewrite the root
        root = self.handle(model.root, ctx)

        # the new body contains the extracted top level logicals and maybe the rewritten root
        body = ctx.top_level if root.replacement is None else ctx.top_level + [root.replacement]

        # create the new model, updating relations and root
        return ir.Model(
            model.engines,
            OrderedSet.from_iterable(model.relations).update(ctx.relations).frozen(),
            model.types,
            ir.Logical(model.root.engine, tuple(), tuple(body))
        )


    #--------------------------------------------------
    # Helper Classes
    #--------------------------------------------------

    class Frame():
        """ The scope of a Logical. """
        binders = (ir.Lookup, ir.Not, ir.Exists, ir.ForAll)
        effects = (ir.Update, ir.Output)

        def __init__(self, task: ir.Logical):
            # the set of task a logical adds to scope. These are tasks in the body of a logical
            # that bind variables, so they must be part the conjunction in inner scopes.
            self.scope: list[ir.Task] = []

            # TODO: we are conservatively pushing these nodes down, but we could be more
            # selective if we did a dataflow analysis

            for t in task.body:
                if isinstance(t, Flatten.Frame.binders):
                    self.scope.append(t)

        def updated_scope_binding(self, all_scope_was_bound: Optional[bool], t: ir.Task, scope_was_bound: bool):
            """ Determine the new state of all_scope_was_bound once task t was handled and returned this scope_was_bound. """
            # for nested logicals, update based on what was returned for the logical
            if isinstance(t, ir.Logical):
                return scope_was_bound and (all_scope_was_bound is None or all_scope_was_bound)

            # if the task is not a binder or an effect, no impact in the flag
            if isinstance(t, Flatten.Frame.binders) or isinstance(t, Flatten.Frame.effects):
                return all_scope_was_bound

            # any other task sets the flag to false
            return False

    @dataclass
    class Context():
        # the logicals that will be at the top level at the end of the rewrite
        top_level: list[ir.Logical] = field(default_factory=list)
        # new relations created during the pass
        relations: list[ir.Relation] = field(default_factory=list)
        # the stack of frames
        frames: list[Flatten.Frame] = field(default_factory=list)

        def get_scope(self) -> OrderedSet[ir.Task]:
            """ Get all tasks in scope, from all frames in scope. """
            tasks = ordered_set()
            for frame in self.frames:
                tasks.update(frame.scope)
            return tasks

    @dataclass
    class HandleResult():
        """ The result of the handle methods. """
        # when a task is handled, the replacement for it, if any; if the task is not changed,
        # this can be the task itself; if the task was extracted as a new logical, for
        # for example, this can be a lookup to the connection relationl; if all sub-tasks
        # became logicals with effects, this can be empty.
        replacement: Optional[ir.Task]

        # True if the flattening of the task bound the scope passed to it; this is used to
        # determine if the scope can be removed from the caller, essentially because it was
        # used by all  children.
        scope_was_bound: bool = field(default=False)

    #--------------------------------------------------
    # IR handlers
    #--------------------------------------------------

    def handle(self, task: ir.Task, ctx: Context):
        if isinstance(task, ir.Logical):
            return self.handle_logical(task, ctx)
        elif isinstance(task, ir.Union):
            return self.handle_union(task, ctx)
        elif isinstance(task, ir.Match):
            return self.handle_match(task, ctx)
        else:
            return Flatten.HandleResult(task)

    def handle_logical(self, task: ir.Logical, ctx: Context):

        # create frame to process the logical
        frame = Flatten.Frame(task)
        ctx.frames.append(frame)

        # recursively handle children, collecting the replacements in the body
        body:OrderedSet[ir.Task] = ordered_set()
        all_scope_was_bound = None
        for t in task.body:
            x = self.handle(t, ctx)
            if x.replacement is not None:
                self._extend_body(body, x.replacement)
            all_scope_was_bound = frame.updated_scope_binding(all_scope_was_bound, t, x.scope_was_bound)
        ctx.frames.pop()

        # if the scope was bound to all relevant children, it can be removed from the body
        if all_scope_was_bound:
            body = body - frame.scope

        # all children were extracted
        if not body:
            return Flatten.HandleResult(None)

        # short-circuit when there's a single child that is a logical, return it
        if len(body) == 1 and isinstance(body.some(), ir.Logical):
            t = cast(ir.Logical, body.some())
            engine = task.engine if t.engine is None else t.engine
            return Flatten.HandleResult(ir.Logical(engine, task.hoisted, t.body))

        groups = group_tasks(body.list, {
            "outputs": ir.Output,
            "updates": ir.Update,
            "aggregates": ir.Aggregate,
        })

        # if there are outputs, currently assume it's already at top level, so just return
        # the rewritten body
        if groups["outputs"]:
            return Flatten.HandleResult(ir.Logical(task.engine, task.hoisted, tuple(body)))

        # if there are updates, extract as a new top level rule
        if groups["updates"]:
            # add whatever was in scope at the start of the body
            body = ctx.get_scope() | body
            ctx.top_level.append(ir.Logical(task.engine, task.hoisted, tuple(body)))

            # no need to refer to the extracte logical because it is an update
            return Flatten.HandleResult(None, True)

        if groups["aggregates"]:
            if len(groups["aggregates"]) > 1:
                # stop rewritting as we don't know how to handle this yet
                return Flatten.HandleResult(task)

            # there must be only one
            agg = cast(ir.Aggregate, groups["aggregates"].some())

            # add whatever was in scope at the start of the body
            body = ctx.get_scope() | body

            # extract a new logical for the aggregate
            exposed_vars = cast(list[ir.Var], list(agg.group) + [agg.args[-1]])
            connection = self._extract(agg, body, exposed_vars, ctx)

            # return a reference to the connection relation
            reference = f.logical([f.lookup(connection, exposed_vars)], self._merge_var_list(exposed_vars, task.hoisted))
            return Flatten.HandleResult(reference, True)

        return Flatten.HandleResult(ir.Logical(task.engine, task.hoisted, tuple(body)))


    def handle_match(self, match: ir.Match, ctx: Context):
        # TODO: how to deal with malformed input like this?
        if not match.tasks:
            return Flatten.HandleResult(match)

        body = ctx.get_scope()
        exposed_vars = visitor.collect_vars(*body)
        exposed_vars.update(metamodel_utils.hoisted_vars(match.hoisted))
        exposed_vars = exposed_vars.list

        connection = None
        reference = None

        for branch in match.tasks:
            # process the branch
            x = self.handle(branch, ctx)
            assert(x.replacement)

            branch_body: OrderedSet[ir.Task] = OrderedSet.from_iterable(body)
            self._extend_body(branch_body, x.replacement)
            if reference:
                branch_body.add(self._negate(reference, len(match.hoisted)))
                branch_body = OrderedSet.from_iterable([f.union([f.logical(branch_body.list), reference])])
            connection = self._extract(branch, branch_body, exposed_vars, ctx, "match")
            reference = f.logical([f.lookup(connection, exposed_vars)], exposed_vars)

        return Flatten.HandleResult(reference, True)


    def handle_union(self, union: ir.Union, ctx: Context):
        # TODO: how to deal with malformed input like this?
        if not union.tasks:
            return Flatten.HandleResult(union)

        body = ctx.get_scope()
        exposed_vars = visitor.collect_vars(*body)
        exposed_vars.update(metamodel_utils.hoisted_vars(union.hoisted))
        exposed_vars = exposed_vars.list

        connection = None

        for branch in union.tasks:
            # process the branch
            x = self.handle(branch, ctx)
            assert(x.replacement)

            branch_body: OrderedSet[ir.Task] = OrderedSet.from_iterable(body)
            self._extend_body(branch_body, x.replacement)

            if connection is None:
                # first branch, extract making a connection relation
                connection = self._extract(branch, branch_body, exposed_vars, ctx, "union")
            else:
                # subsequent branch, extract reusing the connection relation
                # add derivation to the extracted body
                branch_body.add(f.derive(connection, exposed_vars))

                # extract the body
                ctx.top_level.append(ir.Logical(union.engine, tuple(), tuple(branch_body)))

        # return a reference to the connection
        assert(connection)
        reference = f.logical([f.lookup(connection, exposed_vars)], exposed_vars)
        return Flatten.HandleResult(reference, True)

    #--------------------------------------------------
    # Helpers
    #--------------------------------------------------

    def _negate(self, reference: ir.Logical, values: int):
        """
        Return a negation of this reference, where the last `values` arguments are to
        be replaced by wildcards (i.e. len(reference.args) - values are keys so they need
        to be bound in the Not.)
        """
        lookup = cast(ir.Lookup, reference.body[0])
        args = []
        i = 0
        last = len(lookup.args) - values
        for arg in lookup.args:
            args.append(f.wild()) if i >= last else args.append(arg)
            i += 1

        return ir.Not(reference.engine, f.lookup(lookup.relation, args))


    def _extract(self, task: ir.Task, body: OrderedSet[ir.Task], exposed_vars: list[ir.Var], ctx: Context, prefix: Optional[str]=None) -> ir.Relation:
        """
        Extract into this context a new top level Logical that contains this body plus a
        derive task into a new temporary relation, which is also registered with the ctx.
        The exposed_vars determine the arguments of this temporary relation. The prefix
        can be used to customize the name of the relation, which defaults to the task kind.

        Return the temporary relation created for the extraction.
        """

        # TODO: review variable rewrites, i.e. when we extract a new logical, we should remap variables
        p = prefix if prefix else task.kind

        # new relation to derive the aggregation into
        connection = f.relation(f"_{p}_{task.id}", [f.field(v.name, v.type) for v in exposed_vars])
        ctx.relations.append(connection)

        # add derivation to the extracted body
        body.add(f.derive(connection, exposed_vars))

        # extract the body
        ctx.top_level.append(ir.Logical(task.engine, tuple(), tuple(body)))

        # return a reference to the connection relation
        return connection


    def _merge_var_list(self, vars: list[ir.Var], hoisted: tuple[ir.VarOrDefault, ...]) -> list[ir.VarOrDefault]:
        """ Merge vars and hoisted, making sure that hoisted vars have precedence since they may have defaults. """
        r = []
        hoisted_vars = metamodel_utils.hoisted_vars(hoisted)
        for v in vars:
            if v not in hoisted_vars:
                r.append(v)
        r.extend(hoisted)
        return r

    def _extend_body(self, body: OrderedSet[ir.Task], extra: ir.Task):
        """ Add the extra task to the body, but if the extra is a simple logical, just
        inline its subtasks. """
        if isinstance(extra, ir.Logical):
            if extra.hoisted:
                # hoists, remove things that are already in the body to avoid duplicates
                logical_body = []
                for t in extra.body:
                    if t not in body:
                        logical_body.append(t)
                if len(logical_body) == len(extra.body):
                    # no duplicates
                    body.add(extra)
                else:
                    # some duplicate, remove them
                    body.add(ir.Logical(
                        extra.engine,
                        extra.hoisted,
                        tuple(logical_body)
                    ))
            else:
                # no hoists, just inline
                body.update(extra.body)
        else:
            body.add(extra)



"""
    Elementary IR relations.
"""
import sys

from . import ir, factory as f
from . import types

#
# Relations
#

# Comparators
def _comparator(name: str):
    return f.relation(name, [f.input_field("a", types.Number), f.input_field("b", types.Number)])

gt = _comparator(">")
gte = _comparator(">=")
lt = _comparator("<")
lte = _comparator("<=")
neq = _comparator("!=")
eq = _comparator("=")

# Arithmetic operators
def _binary_op(name: str):
    return f.relation(name, [f.input_field("a", types.Number), f.input_field("b", types.Number), f.field("c", types.Number)])

plus = _binary_op("+")
minus = _binary_op("-")
div = _binary_op("/")
mul = _binary_op("*")
mod = _binary_op("%")

abs = f.relation("abs", [f.input_field("a", types.Number), f.field("b", types.Number)])

# Other
# TODO careful with range - unlike rel, python stops at stop-1
range = f.relation("range", [
    f.input_field("start", types.Number),
    f.input_field("stop", types.Number),
    f.input_field("step", types.Number),
    f.field("result", types.NumberSet),
])
concat = f.relation("concat", [f.input_field("a", types.String), f.input_field("b", types.String), f.field("c", types.String)])
hash = f.relation("rel_primitive_hash_tuple", [f.input_field("args", types.AnyList), f.field("hash", types.Hash)])

solverlib_fo_appl = f.relation("rel_primitive_solverlib_fo_appl", [
    f.input_field("op", types.Int),
    f.input_field("args", types.AnyList),
    f.field("result", types.String),
])

#
# Annotations
#

# indicates a relation is external to the system and, thus, backends should not rename or
# otherwise modify it.
external = f.relation("external", [])
# reusable annotation object.
external_annotation = f.annotation(external, [])

#
# Aggregations
#
def aggregation(name: str, params: list[ir.Field]):
    """Defines an aggregation, which is a Relation whose first 2 fields are a projection
    and a group, followed by the params."""
    fields = [
        f.input_field("projection", types.AnyList),
        f.input_field("group", types.AnyList)
    ] + params
    return f.relation(name, fields)

concat = aggregation("concat", [
    f.input_field("sep", types.String),
    f.input_field("over", types.StringSet),
    f.field("result", types.String)
])
# note that count does not need "over" because it counts the projection
count = aggregation("count", [
    f.field("result", types.Number)
])
stats = aggregation("stats", [
    f.input_field("over", types.NumberSet),
    f.field("std_dev", types.Number),
    f.field("mean", types.Number),
    f.field("median", types.Number),
])
sum = aggregation("sum", [
    f.input_field("over", types.NumberSet),
    f.field("result", types.Number)
])
avg = aggregation("avg", [
    f.input_field("over", types.NumberSet),
    f.field("result", types.Number)
])
max = aggregation("max", [
    f.input_field("over", types.NumberSet),
    f.field("result", types.Number)
])
min = aggregation("min", [
    f.input_field("over", types.NumberSet),
    f.field("result", types.Number)
])


# TODO: these are Rel specific, should be moved from here
# Conversions
string = f.relation("string", [f.input_field("a", types.Any), f.field("b", types.String)])
parse_date = f.relation("parse_date", [f.input_field("a", types.String), f.input_field("b", types.String), f.field("c", types.Number)])
parse_datetime = f.relation("parse_datetime", [f.input_field("a", types.String), f.input_field("b", types.String), f.field("c", types.Number)])
parse_decimal = f.relation("parse_decimal", [f.input_field("a", types.Int), f.input_field("b", types.Int), f.input_field("c", types.String), f.field("d", types.Decimal)])

solverlib_ho_appl = aggregation("rel_primitive_solverlib_ho_appl", [
    f.input_field("op", types.Int),
    f.field("result", types.String),
])

#
# Public ccess to built-in relations
#

def is_builtin(r: ir.Relation):
    return r in builtin_relations

def is_annotation(r: ir.Relation):
    return r in builtin_annotations

def _compute_builtin_relations() -> list[ir.Relation]:
    module = sys.modules[__name__]
    relations = []
    for name in dir(module):
        builtin = getattr(module, name)
        if isinstance(builtin, ir.Relation) and builtin not in builtin_annotations:
            relations.append(builtin)
    return relations

# manually maintain the list of relations that are actually annotations
builtin_annotations = [external]
builtin_annotations_by_name = dict((r.name, r) for r in builtin_annotations)

builtin_relations = _compute_builtin_relations()
builtin_relations_by_name = dict((r.name, r) for r in builtin_relations)



from collections import defaultdict

from relationalai.early_access.dsl.adapters.orm.parser import ORMParser
from relationalai.early_access.dsl.core.types.standard import Integer, UnsignedInteger, String, DateTime, Date, Decimal
from relationalai.early_access.dsl.ontologies.models import Model
from relationalai.early_access.dsl.types.entities import AbstractEntityType


class ORMAdapter:

    def __init__(self, orm_file_path: str):
        self._parser = ORMParser(orm_file_path)
        self._relationship_role_value_constraints = defaultdict()
        self.model = self.orm_to_model()

    def orm_to_model(self):
        model = Model(self._parser.model_name())

        self._add_value_types(model)
        self._add_entity_types(model)
        self._add_subtype_relationships(model)
        self._add_relationships(model)
        self._add_external_identifying_relationships(model)
        self._add_role_value_constraints(model)

        return model

    def _add_value_types(self, md):
        for vt in self._parser.value_types().values():
            md.value_type(vt.name, self._map_datatype(vt.data_type))

    def _add_entity_types(self, md):
        for et in self._parser.entity_types().values():
            md.entity_type(et.name)

    def _add_subtype_relationships(self, md):
        for parent, children in self._parser.subtype_facts().items():
            supertype = md.lookup_concept(parent)
            subtypes = []
            for child in children:
                sub = md.lookup_concept(child)
                subtypes.append(sub)
            md.subtype_arrow(supertype, subtypes)

    def _add_relationships(self, md):
        object_types = self._parser.object_types()
        role_value_constraints = self._parser.role_value_constraints()
        internal_uniqueness_constraints = self._parser.internal_uniqueness_constraints()
        fact_type_to_roles = self._parser.fact_type_to_roles()
        fact_type_to_unique_constraints = self._parser.fact_type_to_unique_constraints()
        for fact_type, reading_orders in self._parser.fact_type_readings().items():
            with (md.relationship() as rel):
                role_idx_to_player = list()
                # Use the first reading to add the players
                for role in fact_type_to_roles[fact_type]:
                    role_name = role.name if role.name != "" else None
                    p = md.lookup_concept(object_types[role.player].name)
                    rel.role(p, name=role_name)
                    role_idx_to_player.append(p)
                    if role.id in role_value_constraints:
                        self._relationship_role_value_constraints[rel] = role_value_constraints[role.id]
                # Create the readings
                for rdo in reading_orders:
                    argz = []
                    for player, txt in zip(rdo["players"], rdo["texts"]):
                        p = md.lookup_concept(object_types[player].name)
                        role_idx = role_idx_to_player.index(p)
                        rl = rel.role_at(role_idx)
                        argz.append(rl)
                        if txt != "":
                            argz.append(txt)
                    rel.relation(*argz)
                # Marking identifying relationships
                if fact_type in fact_type_to_unique_constraints:
                    for uc_id in fact_type_to_unique_constraints[fact_type]:
                        uc = internal_uniqueness_constraints[uc_id]
                        if uc.identifies:
                            et = md.lookup_concept(object_types[uc.identifies].name)
                            role_name = role_idx_to_player.index(et)
                            md.ref_scheme(rel.relations()[role_name])

    def _add_external_identifying_relationships(self, model):
        roles = self._parser.roles()
        object_types = self._parser.object_types()
        for uc_id, uc in self._parser.external_uniqueness_constraints().items():
            et = model.lookup_concept(object_types[uc.identifies].name)
            identifying_rel = []
            for ro in uc.roles:
                relationship = model.lookup_relationship(roles[ro].relationship_name)
                for rl in relationship.roles():
                    if rl.player_type == et:
                        idx = relationship.roles().index(rl)
                        identifying_rel.append(relationship.relations()[idx])
            model.ref_scheme(*identifying_rel)

    def _add_role_value_constraints(self, model):
        for rel, values in self._relationship_role_value_constraints.items():
            for rls in rel.relations():
                if isinstance(rls.reading().role_at(0).player(), AbstractEntityType):
                    model.role_value_constraint(rls, values=values)
                    break

    @staticmethod
    def _map_datatype(dt):
        mapping = {
            "AutoCounterNumericDataType": Integer,
            "UnsignedIntegerNumericDataType": UnsignedInteger,
            "VariableLengthTextDataType": String,
            "DateAndTimeTemporalDataType": DateTime,
            "DecimalNumericDataType": Decimal,
            "DateTemporalDataType": Date
        }
        return mapping.get(dt, String)


from __future__ import annotations
from dataclasses import dataclass, field
from typing import Tuple, Union
from typing import cast

from relationalai.early_access.metamodel import ir, compiler as c, factory as f, types, visitor
from relationalai.early_access.metamodel.util import FrozenOrderedSet, OrderedSet, ordered_set, group_by, split_by

class Denormalize(c.Pass):
    """
    Pass to denormalize relations, grouping relations that have the same entity as key.

    General algorithm:

    1. when handling the Model, go to its relations and find the ones that have common
    entity keys. They will be denormalized (joined together as a single "table").
    2. when handling lookups and updates that refer to one of these relations that were
    denormalized, change the reference to be to the new relation and fill the blanks with
    Nones, which represent NULLs.
    3. finally, when handling Logical nodes, try to group together lookups and updates that
    are refering to the same denormalized relation. This may remove some of the NULLs.
    """
    def rewrite(self, model: ir.Model, cache) -> ir.Model:
        return OldDenormalize().walk(model)

@dataclass
class OldDenormalize(visitor.DeprecatedPass):
    # TODO: use the new Rewriter when available.

    denormalized: dict[ir.Relation, ir.Relation] = field(default_factory=dict, init=False, hash=False, compare=False)

    def reset(self):
        self.context = dict()

    def handle_model(self, model: ir.Model, parent: None):
        # denormalize some relations in the model, set it in the context, and then push
        # that context to the root traversal so that references to the denormalized
        # relations are adjusted.

        # denormalize the relations
        relations, denormalized = self._denormalize_relations(model.relations)

        # rewrite tasks to refer to denormalized relations
        return ir.Model(
            self.walk_set(model.engines, model),
            relations,
            self.walk_set(model.types, model),
            self.walk(model.root, model),
        )

    def handle_update(self, node: ir.Update, parent: ir.Node):
        # if the relation to be updated was denormalized, denormalize the reference

        if node.relation not in self.denormalized:
            return super().handle_update(node, parent)

        denormalized_relation = self.denormalized[node.relation]
        args = self._denormalize_reference(
            denormalized_relation, node.relation, node.args
        )

        return ir.Update(
            node.engine,
            denormalized_relation,
            args,
            node.effect,
            node.annotations
        )

    def handle_lookup(self, node: ir.Lookup, parent: ir.Node):
        # if the relation to be looked up was denormalized, denormalize the reference

        if node.relation not in self.denormalized:
            return super().handle_lookup(node, parent)

        denormalized_relation = self.denormalized[node.relation]
        args = self._denormalize_reference(
            denormalized_relation, node.relation, node.args
        )

        return ir.Lookup(
            node.engine,
            denormalized_relation,
            args
        )

    def handle_logical(self, node: ir.Logical, parent: ir.Node):
        # denormalize the body first
        body = self.walk_list(node.body, node)

        # function to compute a key to group Lookups or Updates that can be merged
        def task_key(t: ir.Task):
            assert isinstance(t, (ir.Lookup, ir.Update))
            return (t.kind, t.relation, t.args[0])

        # now group references
        tasks, new_body = split_by(body, lambda t: isinstance(t, (ir.Lookup, ir.Update)))
        for _, grouped_tasks in group_by(tasks, task_key).items():
            grouped_tasks = cast(OrderedSet[Union[ir.Lookup, ir.Update]], grouped_tasks)

            # get some of the tasks as a representative
            some = grouped_tasks.some()

            if len(grouped_tasks) == 1:
                new_body.append(some)
                continue

            # now join the arguments of the multiple references
            # TODO: this will fail if we have multiple tasks on the same arg
            new_args = []
            i = 0
            for _ in some.relation.fields:
                # note that we are still using None to indicate we don't care about the column
                v = None
                for t in grouped_tasks:
                    if t.args[i] is not None:
                        v = t.args[i]
                        break
                new_args.append(v)
                i += 1

            # finally, create a rewritten task for the whole group
            if isinstance(some, ir.Lookup):
                new_body.append(ir.Lookup(some.engine, some.relation, tuple(new_args)))
            else:
                new_body.append(ir.Update(some.engine, some.relation, tuple(new_args), some.effect, some.annotations)) # TODO maybe merge annos?
        return ir.Logical(
            node.engine,
            node.hoisted,
            tuple(new_body)
        )


    def _denormalize_reference(self, denormalized_relation, relation, args) -> Tuple[ir.Value, ...]:
        """ Adjust arguments to a denormalized reference.

            A reference to `relation` was used with `args`. But `relation` was denormalized
            into `denormalized_relation`. This method returns a new tuple of adjusted args
            to account for the denormalization, filling Nones as appropriate.
        """

        # the first arg is always the entity
        new_args = [args[0]]
        # for the rest of the denormalized relation columns, lookup only if the field came
        # from this relation, otherwise use None
        i = 1
        for fld in denormalized_relation.fields[1:]:
            if fld in relation.fields:
                new_args.append(args[i])
                i += 1
            else:
                new_args.append(None)
        return tuple(new_args)

    def _denormalize_relations(self, relations: FrozenOrderedSet[ir.Relation]) -> \
        Tuple[FrozenOrderedSet[ir.Relation], dict[ir.Relation, ir.Relation]]:

        """ Denormalize the relations that can be denormalized.

        Group together relations that are keyed by the same "entity". This method defines
        entities as being types that haver a unary relation containing only that type. All
        relations whose first argument is this type are grouped together.

        Returns a tuple with 2 elements:
        1. the new set of relations after denormalization
        2. a dict from relations that were denormalized away to the new relation that took
        its place.
        """
        new_relations = ordered_set()
        denormalized: dict[ir.Relation, ir.Relation] = dict()

        # entities are non-builtin types that have a unary relation
        entity_types: OrderedSet[ir.Type] = ordered_set()
        entity_relations: dict[ir.Type, ir.Relation] = dict()

        for r in relations:
            if len(r.fields) == 1 and not types.is_builtin(r.fields[0].type):
                e = r.fields[0].type
                entity_types.add(e)
                entity_relations[e] = r

        # filter out relations that don't have an entity as a key
        with_entity, unaffected = split_by(
            relations,
            lambda r:
                r.fields[0].type in entity_types and  # the first element is an entity
                all(not isinstance(f.type, ir.SetType) for f in r.fields) # no field is multi
                # TODO - we need to account for wide relations that have multis
        )
        new_relations.update(unaffected)

        # group relations by entity, they will be denormalized
        by_entity = group_by(with_entity, lambda r: cast(ir.ScalarType, r.fields[0].type))
        for e_type, group in by_entity.items():
            r = entity_relations[e_type]
            fields = []
            # the first column of the denormalized relation is the "id"
            fields.append(r.fields[0])
            for relation in group:
                # add all fields except the first entity field
                # TODO - we may have multiple relations with the same name
                fields.extend(relation.fields[1:])

            # create the new denormalized relation, add it to the new_relations set
            denormalized_relation = f.relation(
                e_type.name,
                fields
            )
            new_relations.add(denormalized_relation)

            # make sure the original relations point to the new relation
            for g in group:
                denormalized[g] = denormalized_relation

        return new_relations.frozen(), denormalized



from __future__ import annotations
import importlib.metadata

from typing import cast

from .clients import config as cfg
from .clients.config import Config, save_config
from . import dsl
from . import debugging
from . import metamodel
from . import rel
from .loaders import csv
from . import analysis
from . import tools
from snowflake.snowpark import Session
from .errors import RAIException, handle_missing_integration
from .environments import runtime_env, SessionEnvironment
import __main__

__version__ = importlib.metadata.version(__package__ or __name__)

def Model(
    name: str,
    *,
    profile: str | None = None,
    config: Config | None = None,
    dry_run: bool | None = False,
    debug: bool | None = None,
    debug_host: str | None = None,
    debug_port: int | None = None,
    connection: Session | None = None,
    keep_model: bool | None = None,
    isolated: bool | None = None,
    nowait_durable: bool | None = None,
    use_package_manager: bool | None = None,
    ensure_change_tracking: bool | None = None,
    format: str = "default",
):
    config = config or Config(profile=profile)
    if use_package_manager is not None:
        config.set("use_package_manager", use_package_manager)
    if ensure_change_tracking is not None:
        config.set("ensure_change_tracking", ensure_change_tracking)

    if isinstance(runtime_env, SessionEnvironment):
        connection = runtime_env.configure_session(config, connection)

    if debug is None:
        config_debug = config.get("debug", True)
        if isinstance(config_debug, dict):
            debug = True
        elif isinstance(config_debug, bool):
            debug = config_debug
        else:
            raise Exception("Invalid value specified for `debug`, expected `true` or `false`.")

    if debug_host is None:
        # Our get function isn't robust to allowing `debug = true/false` or `[debug]\n  port=...`
        # Went with the lowest impact solve for now which is handling it locally.
        try:
            debug_host = config.get("debug.host", None)
        except AttributeError:
            pass

    if debug_port is None:
        try:
            config_debug_port = config.get("debug.port", 8080)
            if not isinstance(config_debug_port, int):
                raise Exception("Invalid value specified for `debug.port`, expected `int`.")
            debug_port = config_debug_port
        except AttributeError:
            pass

    if debug and not runtime_env.remote:
        from relationalai.tools.debugger_client import start_debugger_session
        start_debugger_session(config, host=debug_host, port=debug_port)

    main_path = getattr(__main__, "__file__", None)
    debugging.create_program_span_if_not_exists(main_path, config)

    if not config.file_path:
        if cfg.legacy_config_exists():
            message = (
                "Use `rai init` to migrate your configuration file "
                "to the new format (raiconfig.toml)"
            )
        else:
            message = "No configuration file found. Please run `rai init` to create one."
        raise Exception(message)
    if config.get("platform", None) is None:
        config.set("platform", "snowflake")
    platform = config.get("platform")
    if platform != "snowflake" and connection is not None:
        raise ValueError("The `connection` parameter is only supported with the Snowflake platform")
    if dry_run is None:
        dry_run = config.get_bool("compiler.dry_run", False)
    if keep_model is None:
        keep_model = config.get_bool("model.keep", False)
    if isolated is None:
        isolated = config.get_bool("model.isolated", True)
    if nowait_durable is None:
        nowait_durable = config.get_bool("model.nowait_durable", True)

    try:
        if platform == "azure":
            import relationalai.clients.azure as azure
            model = azure.Graph(
                name,
                profile=profile,
                config=config,
                dry_run=dry_run,
                isolated=isolated,
                keep_model=keep_model,
                format=format,
            )
        elif platform == "snowflake":
            from relationalai.clients import snowflake
            model = snowflake.Graph(
                name,
                profile=profile,
                config=config,
                dry_run=dry_run,
                isolated=isolated,
                connection=connection,
                keep_model=keep_model,
                nowait_durable=nowait_durable,
                format=format,
            )
        else:
            raise Exception(f"Unknown platform: {platform}")
    except RAIException as e:
        raise e.clone() from None
    except Exception as e:
        handle_missing_integration(e)
        raise e
    return model


def Resources(
    profile: str | None = None,
    config: Config | None = None,
    connection: Session | None = None,
):
    config = config or Config(profile)
    platform = config.get("platform", "snowflake")
    if platform == "azure":
        from relationalai.clients.azure import Resources
        return Resources(config=config)
    elif platform == "snowflake":
        from relationalai.clients.snowflake import Resources
        return Resources(config=config, connection=connection)
    else:
        raise Exception(f"Unknown platform: {platform}")

def Provider(
    profile: str | None = None,
    config: Config | None = None,
    connection: Session | None = None,
):
    resources = Resources(profile, config, connection)
    platform = resources.config.get("platform", "snowflake")
    if platform == "azure":
        import relationalai.clients.azure
        resources = cast(relationalai.clients.azure.Resources, resources)
        return relationalai.clients.azure.Provider(
            resources=resources
        )
    elif platform == "snowflake":
        import relationalai.clients.snowflake
        resources = cast(relationalai.clients.snowflake.Resources, resources)
        return relationalai.clients.snowflake.Provider(
            resources=resources,
        )
    else:
        raise Exception(f"Unknown platform: {platform}")


def Graph(name:str, dry_run:bool=False):
    return Model(name, profile=None, dry_run=dry_run)

__all__ = [
    'Model', 'Config', 'Resources', 'Provider', 'dsl', 'rel', 'debugging', 'metamodel', 'csv',
    'analysis', 'tools', 'save_config'
]



from __future__ import annotations

from typing import Optional, Sequence as PySequence, cast

from relationalai.early_access.metamodel import ir, compiler as c, factory as f, types, util
from relationalai.early_access.metamodel.visitor import collect_implicit_vars

class Splinter(c.Pass):
    """
    Splits multi-headed rules into multiple rules. Additionally, infers missing Exists tasks.
    """
    def rewrite(self, model: ir.Model, cache) -> ir.Model:
        if isinstance(model.root, ir.Logical):
            final = []
            new_relations:list[ir.Relation] = []
            new_relations.extend(model.relations)
            for child in model.root.body:
                new_logicals, relation = self.split(cast(ir.Logical, child))
                final.extend(new_logicals)
                if relation:
                    new_relations.append(relation)
            return ir.Model(
                    model.engines,
                    util.FrozenOrderedSet.from_iterable(new_relations),
                    model.types,
                    ir.Logical(
                        model.root.engine,
                        model.root.hoisted,
                        tuple(final)
                    )
                )
        return model

    def split(self, node: ir.Logical) -> tuple[list[ir.Logical], Optional[ir.Relation]]:
        # Split this logical, which represents a rule, into potentially many logicals, one
        # for each head (update or output)
        effects, body = self.split_items(node.body)
        if not body:
            return [node], None

        effects_vars = collect_implicit_vars(*effects)
        implicit_vars = list(collect_implicit_vars(*body) - effects_vars)
        has_aggregate = any(isinstance(t, ir.Aggregate) for t in body)

        if len(effects) > 1:
            connection = f.relation(f"q{node.id}", [f.field("", types.Any) for v in effects_vars])
            if implicit_vars:
                args = [f.exists(implicit_vars, f.logical(body))]
            else:
                args = body
            final:list[ir.Logical] = [f.logical([*args, f.derive(connection, list(effects_vars))])]
            for effect in effects:
                effect_vars = collect_implicit_vars(effect)
                lookup_vars = [(v if v in effect_vars else f.wild()) for v in effects_vars]
                final.append(f.logical([f.lookup(connection, lookup_vars), effect]))
            return final, connection

        if has_aggregate and implicit_vars:
            # the implicit_vars must be in a new logical at the same level as the aggregations,
            # so that they are pushed inside the body
            aggs = util.filter_by_type(body, ir.Aggregate)
            for a in aggs:
                body.remove(a)
            return [f.logical([
                f.logical([
                    f.logical([
                        f.exists(implicit_vars, f.logical(body))
                    ]),
                   *aggs
                ]),
                *effects
            ])], None
        elif implicit_vars:
            # add an exists for the implicit vars
            return [f.logical([f.exists(implicit_vars, f.logical(body)), *effects])], None
        else:
            return [node], None


    def split_items(self, items: PySequence[ir.Task]) -> tuple[list[ir.Task], list[ir.Task]]:
        effects = []
        body = []
        for item in items:
            if isinstance(item, (ir.Update, ir.Output)):
                effects.append(item)
            else:
                body.append(item)
        return effects, body



from __future__ import annotations

import duckdb
from pandas import DataFrame
from typing import Any, Union

from relationalai.early_access.metamodel import ir, executor as e
from . import Compiler

class Executor(e.Executor):

    def execute(self, model: ir.Model, task:ir.Task) -> Union[DataFrame, Any]:
        c = Compiler()
        sql = c.compile(model)
        connection = duckdb.connect()
        try:
            result = connection.query(sql).to_df()
            return result
        finally:
            connection.close()



import textwrap
from typing import List, Optional

from pandas import DataFrame
import relationalai as rai
from relationalai import debugging
from relationalai.clients import result_helpers
from relationalai.early_access.dsl.ir.compiler import Compiler
from relationalai.early_access.dsl.ontologies.models import Model
from relationalai.early_access.metamodel import ir
from relationalai.early_access.rel.compiler import ModelToRel


class RelExecutor:

    def __init__(self, database: str, dry_run: bool = False) -> None:
        super().__init__()
        self.database = database
        self.dry_run = dry_run
        self.compiler = Compiler()
        self.model_to_rel = ModelToRel()
        self._resources = None
        self._last_model = None

    @property
    def resources(self):
        if not self._resources:
            with debugging.span("create_session"):
                self._resources = rai.clients.snowflake.Resources()
                self._resources.config.set("use_graph_index", False)
                if not self.dry_run:
                    try:
                        if not self._resources.get_database(self.database):
                            self._resources.create_graph(self.database)
                    except Exception as e:
                        if "already exists" not in str(e).lower():
                            raise e
                    self.engine = self._resources.config.get("engine", strict=False)
                    if not self.engine:
                        self.engine = self._resources.get_default_engine_name()
        return self._resources

    def execute_model(self, model: Model, result_cols: Optional[List[str]] = None) -> DataFrame:
        ir_model = self.compiler.compile_model(model)
        query_ir_model = self.compiler.compile_queries(model.queries())
        return self.execute(ir_model, query_ir_model, result_cols)

    def execute(self, model: ir.Model, query: ir.Model, result_cols: Optional[List[str]] = None) -> DataFrame:
        resources = self.resources

        rules_code = ""
        if self._last_model != model:
            with debugging.span("compile", metamodel=model) as install_span:
                rules_code = str(self.model_to_rel.to_rel(model))
                install_span["compile_type"] = "model"
                install_span["rel"] = rules_code
                rules_code = resources.create_models_code([("pyrel_qb_0", rules_code)])
                self._last_model = model

        with debugging.span("compile", metamodel=query) as compile_span:
            query_code = str(self.model_to_rel.to_rel(query, options={"no_declares": True}))
            compile_span["compile_type"] = "query"
            compile_span["rel"] = query_code

        full_code = textwrap.dedent(f"""
            {rules_code}
            {query_code}
        """)

        if self.dry_run:
            return DataFrame()

        raw_results = self.resources.exec_raw(self.database, self.engine, full_code, False, nowait_durable=True)
        df, _ = result_helpers.format_results(raw_results, None, result_cols)  # Pass None for task parameter
        return df


from abc import abstractmethod

from relationalai.early_access.dsl.core.types import Type

class BindableTable:
    """
    A class representing a bindable table.
    """

    @abstractmethod
    def key_type(self) -> Type:
        pass

    @abstractmethod
    def physical_name(self) -> str:
        pass


class BindableAttribute:

    @property
    @abstractmethod
    def table(self) -> 'BindableTable':
        pass

    @abstractmethod
    def physical_name(self) -> str:
        pass

    @abstractmethod
    def type(self) -> 'Type':
        pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: relationalai/lqp/v1/logic.proto
# Protobuf Python Version: 4.25.6
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1frelationalai/lqp/v1/logic.proto\x12\x13relationalai.lqp.v1\"\x80\x01\n\x0b\x44\x65\x63laration\x12,\n\x03\x64\x65\x66\x18\x01 \x01(\x0b\x32\x18.relationalai.lqp.v1.DefH\x00R\x03\x64\x65\x66\x12/\n\x04loop\x18\x02 \x01(\x0b\x32\x19.relationalai.lqp.v1.LoopH\x00R\x04loopB\x12\n\x10\x64\x65\x63laration_type\"\xa6\x01\n\x03\x44\x65\x66\x12\x33\n\x04name\x18\x01 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\x04name\x12\x34\n\x04\x62ody\x18\x02 \x01(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x04\x62ody\x12\x34\n\x05\x61ttrs\x18\x03 \x03(\x0b\x32\x1e.relationalai.lqp.v1.AttributeR\x05\x61ttrs\"\xaf\x01\n\x04Loop\x12\x41\n\x0ctemporal_var\x18\x01 \x01(\x0b\x32\x1e.relationalai.lqp.v1.LoopIndexR\x0btemporalVar\x12.\n\x05inits\x18\x02 \x03(\x0b\x32\x18.relationalai.lqp.v1.DefR\x05inits\x12\x34\n\x04\x62ody\x18\x03 \x03(\x0b\x32 .relationalai.lqp.v1.DeclarationR\x04\x62ody\"o\n\x0b\x41\x62straction\x12,\n\x04vars\x18\x01 \x03(\x0b\x32\x18.relationalai.lqp.v1.VarR\x04vars\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32\x1c.relationalai.lqp.v1.FormulaR\x05value\"\x8a\x05\n\x07\x46ormula\x12\x35\n\x06\x65xists\x18\x01 \x01(\x0b\x32\x1b.relationalai.lqp.v1.ExistsH\x00R\x06\x65xists\x12\x35\n\x06reduce\x18\x02 \x01(\x0b\x32\x1b.relationalai.lqp.v1.ReduceH\x00R\x06reduce\x12\x44\n\x0b\x63onjunction\x18\x03 \x01(\x0b\x32 .relationalai.lqp.v1.ConjunctionH\x00R\x0b\x63onjunction\x12\x44\n\x0b\x64isjunction\x18\x04 \x01(\x0b\x32 .relationalai.lqp.v1.DisjunctionH\x00R\x0b\x64isjunction\x12,\n\x03not\x18\x05 \x01(\x0b\x32\x18.relationalai.lqp.v1.NotH\x00R\x03not\x12,\n\x03\x66\x66i\x18\x06 \x01(\x0b\x32\x18.relationalai.lqp.v1.FFIH\x00R\x03\x66\x66i\x12/\n\x04\x61tom\x18\x07 \x01(\x0b\x32\x19.relationalai.lqp.v1.AtomH\x00R\x04\x61tom\x12\x35\n\x06pragma\x18\x08 \x01(\x0b\x32\x1b.relationalai.lqp.v1.PragmaH\x00R\x06pragma\x12>\n\tprimitive\x18\t \x01(\x0b\x32\x1e.relationalai.lqp.v1.PrimitiveH\x00R\tprimitive\x12\x36\n\x08true_val\x18\n \x01(\x0b\x32\x19.relationalai.lqp.v1.TrueH\x00R\x07trueVal\x12\x39\n\tfalse_val\x18\x0b \x01(\x0b\x32\x1a.relationalai.lqp.v1.FalseH\x00R\x08\x66\x61lseValB\x0e\n\x0c\x66ormula_type\"j\n\x06\x45xists\x12,\n\x04vars\x18\x01 \x03(\x0b\x32\x18.relationalai.lqp.v1.VarR\x04vars\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32\x1c.relationalai.lqp.v1.FormulaR\x05value\"\xa1\x01\n\x06Reduce\x12\x30\n\x02op\x18\x01 \x01(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x02op\x12\x34\n\x04\x62ody\x18\x02 \x01(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x04\x62ody\x12/\n\x05terms\x18\x03 \x03(\x0b\x32\x19.relationalai.lqp.v1.TermR\x05terms\"?\n\x0b\x43onjunction\x12\x30\n\x04\x61rgs\x18\x01 \x03(\x0b\x32\x1c.relationalai.lqp.v1.FormulaR\x04\x61rgs\"?\n\x0b\x44isjunction\x12\x30\n\x04\x61rgs\x18\x01 \x03(\x0b\x32\x1c.relationalai.lqp.v1.FormulaR\x04\x61rgs\"5\n\x03Not\x12.\n\x03\x61rg\x18\x01 \x01(\x0b\x32\x1c.relationalai.lqp.v1.FormulaR\x03\x61rg\"\x80\x01\n\x03\x46\x46I\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x34\n\x04\x61rgs\x18\x02 \x03(\x0b\x32 .relationalai.lqp.v1.AbstractionR\x04\x61rgs\x12/\n\x05terms\x18\x03 \x03(\x0b\x32\x19.relationalai.lqp.v1.TermR\x05terms\"l\n\x04\x41tom\x12\x33\n\x04name\x18\x01 \x01(\x0b\x32\x1f.relationalai.lqp.v1.RelationIdR\x04name\x12/\n\x05terms\x18\x02 \x03(\x0b\x32\x19.relationalai.lqp.v1.TermR\x05terms\"M\n\x06Pragma\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12/\n\x05terms\x18\x02 \x03(\x0b\x32\x19.relationalai.lqp.v1.TermR\x05terms\"P\n\tPrimitive\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12/\n\x05terms\x18\x02 \x03(\x0b\x32\x19.relationalai.lqp.v1.TermR\x05terms\"\x06\n\x04True\"\x07\n\x05\x46\x61lse\"~\n\x04Term\x12,\n\x03var\x18\x01 \x01(\x0b\x32\x18.relationalai.lqp.v1.VarH\x00R\x03var\x12;\n\x08\x63onstant\x18\x02 \x01(\x0b\x32\x1d.relationalai.lqp.v1.ConstantH\x00R\x08\x63onstantB\x0b\n\tterm_type\"Q\n\x03Var\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x36\n\x04type\x18\x02 \x01(\x0e\x32\".relationalai.lqp.v1.PrimitiveTypeR\x04type\"E\n\x08\x43onstant\x12\x39\n\x05value\x18\x01 \x01(\x0b\x32#.relationalai.lqp.v1.PrimitiveValueR\x05value\"R\n\tAttribute\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x31\n\x04\x61rgs\x18\x02 \x03(\x0b\x32\x1d.relationalai.lqp.v1.ConstantR\x04\x61rgs\"<\n\nRelationId\x12\x15\n\x06id_low\x18\x01 \x01(\x06R\x05idLow\x12\x17\n\x07id_high\x18\x02 \x01(\x06R\x06idHigh\"\x1f\n\tLoopIndex\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\"\x80\x01\n\x0ePrimitiveValue\x12#\n\x0cstring_value\x18\x01 \x01(\tH\x00R\x0bstringValue\x12\x1d\n\tint_value\x18\x02 \x01(\x03H\x00R\x08intValue\x12!\n\x0b\x66loat_value\x18\x03 \x01(\x01H\x00R\nfloatValueB\x07\n\x05value*|\n\rPrimitiveType\x12\x1e\n\x1aPRIMITIVE_TYPE_UNSPECIFIED\x10\x00\x12\x19\n\x15PRIMITIVE_TYPE_STRING\x10\x01\x12\x16\n\x12PRIMITIVE_TYPE_INT\x10\x02\x12\x18\n\x14PRIMITIVE_TYPE_FLOAT\x10\x03\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'relationalai.lqp.v1.logic_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_PRIMITIVETYPE']._serialized_start=2768
  _globals['_PRIMITIVETYPE']._serialized_end=2892
  _globals['_DECLARATION']._serialized_start=57
  _globals['_DECLARATION']._serialized_end=185
  _globals['_DEF']._serialized_start=188
  _globals['_DEF']._serialized_end=354
  _globals['_LOOP']._serialized_start=357
  _globals['_LOOP']._serialized_end=532
  _globals['_ABSTRACTION']._serialized_start=534
  _globals['_ABSTRACTION']._serialized_end=645
  _globals['_FORMULA']._serialized_start=648
  _globals['_FORMULA']._serialized_end=1298
  _globals['_EXISTS']._serialized_start=1300
  _globals['_EXISTS']._serialized_end=1406
  _globals['_REDUCE']._serialized_start=1409
  _globals['_REDUCE']._serialized_end=1570
  _globals['_CONJUNCTION']._serialized_start=1572
  _globals['_CONJUNCTION']._serialized_end=1635
  _globals['_DISJUNCTION']._serialized_start=1637
  _globals['_DISJUNCTION']._serialized_end=1700
  _globals['_NOT']._serialized_start=1702
  _globals['_NOT']._serialized_end=1755
  _globals['_FFI']._serialized_start=1758
  _globals['_FFI']._serialized_end=1886
  _globals['_ATOM']._serialized_start=1888
  _globals['_ATOM']._serialized_end=1996
  _globals['_PRAGMA']._serialized_start=1998
  _globals['_PRAGMA']._serialized_end=2075
  _globals['_PRIMITIVE']._serialized_start=2077
  _globals['_PRIMITIVE']._serialized_end=2157
  _globals['_TRUE']._serialized_start=2159
  _globals['_TRUE']._serialized_end=2165
  _globals['_FALSE']._serialized_start=2167
  _globals['_FALSE']._serialized_end=2174
  _globals['_TERM']._serialized_start=2176
  _globals['_TERM']._serialized_end=2302
  _globals['_VAR']._serialized_start=2304
  _globals['_VAR']._serialized_end=2385
  _globals['_CONSTANT']._serialized_start=2387
  _globals['_CONSTANT']._serialized_end=2456
  _globals['_ATTRIBUTE']._serialized_start=2458
  _globals['_ATTRIBUTE']._serialized_end=2540
  _globals['_RELATIONID']._serialized_start=2542
  _globals['_RELATIONID']._serialized_end=2602
  _globals['_LOOPINDEX']._serialized_start=2604
  _globals['_LOOPINDEX']._serialized_end=2635
  _globals['_PRIMITIVEVALUE']._serialized_start=2638
  _globals['_PRIMITIVEVALUE']._serialized_end=2766
# @@protoc_insertion_point(module_scope)



#--------------------------------------------------
# Compiler
#--------------------------------------------------

from typing import cast, Tuple, Sequence

from relationalai.early_access.metamodel import ir, compiler as c, builtins as rel_builtins
from relationalai.early_access.metamodel import types
from relationalai.early_access.lqp import ir as lqp
from relationalai.early_access.metamodel.rewrite import Splinter
from relationalai.early_access.metamodel.util import NameCache

import hashlib

# TODO: take rewrite out of rel it doesnt belong there
from relationalai.early_access.rel import rewrite

from dataclasses import field

class Compiler(c.Compiler):
    def __init__(self):
        # TODO: are these all needed from the Rel emitter?
        # TODO: should there be a pass to remove aliases from output?
        super().__init__([
            rewrite.ExtractCommonLookups(),
            rewrite.Flatten(),

            # Adds missing existentials + splits multi-headed rules into single rules
            Splinter(),
        ])

    # TODO: what should return value be? string according to parents?
    def do_compile(self, model: ir.Model, options:dict={}) -> str:
        lqp_ir, debugging_ctx = Model2Lqp().to_lqp(model)
        lqp_str = lqp_ir.to_llqp()

        debug_str = ""
        if len(debugging_ctx.id_to_orig_name) > 0:
            debug_str += ";; Original names\n"

        for (rid, name) in debugging_ctx.id_to_orig_name.items():
            debug_str += ";; \t " + str(rid) + " -> `" + name + "`\n"

        if debug_str != "":
            lqp_str += "\n\n"
            lqp_str += ";; Debug information\n"
            lqp_str += ";; -----------------------\n"
            lqp_str += debug_str

        return lqp_str

class DebuggingCtx:
    """ Extra information only used for debugging. """
    def __init__(self, id_to_orig_name: dict[lqp.RelationId, str]):
       self.id_to_orig_name = id_to_orig_name

class Model2Lqp:
    """ Generates LQP IR from the model IR. """
    def __init__(self):
        # TODO: comment htese fields
        # TODO: should we have a pass to rename variables instead of this?
        self.var_name_cache: NameCache = field(default_factory=NameCache)
        self.id_to_orig_name: dict[lqp.RelationId, str] = field(default_factory=dict)
        self.output_ids: list[lqp.RelationId] = field(default_factory=list)

    """ Main access point. Converts the model IR to an LQP program. """
    def to_lqp(self, model: ir.Model) -> Tuple[lqp.LqpProgram, DebuggingCtx]:
        _assert_valid_input(model)
        self._reset_state()
        program = self._translate_to_program(model)
        debugging_ctx = DebuggingCtx(self.id_to_orig_name)
        return (program, debugging_ctx)


    def _reset_state(self):
        self.var_name_cache = NameCache()
        self.id_to_orig_name = {}
        self.output_ids = []

    def _translate_to_program(self, model: ir.Model) -> lqp.LqpProgram:
        decls: list[lqp.Declaration] = []
        outputs: list[Tuple[str, lqp.RelationId]] = []

        # LQP only accepts logical tasks
        # These are asserted at init time
        root = cast(ir.Logical, model.root)
        assert len(root.body) >= 1

        for subtask in root.body:
            # TODO: when do we get more than one?
            child_l = cast(ir.Logical, subtask)
            decl = self._translate_to_decl(child_l)
            decls.append(decl)

        for output_id in self.output_ids:
            assert isinstance(output_id, lqp.RelationId)
            outputs.append(("output", output_id))

        assert len(outputs) >= 1
        return lqp.LqpProgram(decls, outputs)

    def _translate_to_decl(self, rule: ir.Logical) -> lqp.Declaration:
        outputs = []
        updates = []
        body_tasks = []
        aggregates = []

        for task in rule.body:
            if isinstance(task, ir.Output):
                outputs.append(task)
            elif isinstance(task, ir.Lookup):
                body_tasks.append(task)
            elif isinstance(task, ir.Logical):
                body_tasks.append(task)
            elif isinstance(task, ir.Exists):
                body_tasks.append(task)
            elif isinstance(task, ir.Aggregate):
                aggregates.append(task)
            elif isinstance(task, ir.Update):
                updates.append(task)
            else:
                raise NotImplementedError(f"Unknown task type: {type(task)}")

        assert len(aggregates) == 0, "unsupported aggregates yet"

        conjuncts = []
        for task in body_tasks:
            conjunct = self._translate_to_formula(task)
            conjuncts.append(conjunct)
        body = lqp.Conjunction(conjuncts) if len(conjuncts) != 1 else conjuncts[0]

        # TODO: unifying all this?
        if len(outputs) > 0:
            assert len(outputs) == 1, "only one output supported at the moment"
            output = outputs[0]
            assert isinstance(output, ir.Output)
            output_vars = []
            for _, v in output.aliases:
                # TODO: we dont yet handle aliases
                assert isinstance(v, ir.Var)
                var = self._translate_to_var(v)
                output_vars.append(var)
            abstraction = lqp.Abstraction(output_vars, body)
            # TODO: is this correct? might need attrs tooo?
            rel_id = _gen_rel_id(abstraction)
            self.output_ids.append(rel_id)
            return lqp.Def(rel_id, abstraction, [])

        assert len(updates) == 1, "only one update supported at the moment"
        update = updates[0]
        assert isinstance(update, ir.Update)
        effect = update.effect
        assert effect == ir.Effect.derive, "only derive supported at the moment"

        args = []
        for var in update.args:
            assert isinstance(var, ir.Var)
            args.append(self._translate_to_var(var))

        abstraction = lqp.Abstraction(args, body)
        # TODO: is this correct? might need attrs tooo?
        rel_id = _gen_rel_id(abstraction)
        # TODO: attrs?
        self.id_to_orig_name[rel_id] = update.relation.name
        return lqp.Def(rel_id, abstraction, [])

    def _translate_to_formula(self, task: ir.Task) -> lqp.Formula:
        if isinstance(task, ir.Logical):
            conjuncts = []
            for child in task.body:
                conjunct = self._translate_to_formula(child)
                conjuncts.append(conjunct)
            return lqp.Conjunction(conjuncts)
        elif isinstance(task, ir.Lookup):
            return self._translate_to_atom(task)
        elif isinstance(task, ir.Exists):
            lqp_vars = []
            for var in task.vars:
                lqp_vars.append(self._translate_to_var(var))
            formula = self._translate_to_formula(task.task)
            return lqp.Exists(lqp_vars, formula)
        else:
            raise NotImplementedError(f"Unknown task type (formula): {type(task)}")

    def _translate_to_var(self, var: ir.Var) -> lqp.Var:
        name = self.var_name_cache.get_name(var.id, var.name)
        t = type_from_var(var)
        return lqp.Var(name, t)

    def _translate_to_atom(self, task: ir.Lookup) -> lqp.Formula:
        # TODO: want signature not name
        rel_name = task.relation.name
        assert isinstance(rel_name, str)
        terms = []
        sig_types = []
        for arg in task.args:
            if isinstance(arg, lqp.PrimitiveValue):
                term = lqp.Constant(arg)
                terms.append(term)
                t = type_from_constant(arg)
                sig_types.append(t)
                continue
            assert isinstance(arg, ir.Var)
            var = self._translate_to_var(arg)
            terms.append(var)
            sig_types.append(var.type)

        # TODO: wrong
        if rel_builtins.is_builtin(task.relation):
            return self._translate_builtin_to_primitive(task.relation, terms)

        return lqp.RelAtom(lqp.RelationSig(rel_name, sig_types), terms)

    def _translate_builtin_to_primitive(self, relation: ir.Relation, terms: list[lqp.Term]) -> lqp.Primitive:
        lqp_name = self._name_to_lqp_name(relation.name)
        return lqp.Primitive(lqp_name, terms)

    def _name_to_lqp_name(self, name: str) -> str:
        # TODO: do these proprly
        if name == "+":
            return "rel_primitive_add"
        elif name == "-":
            return "rel_primitive_subtract"
        elif name == "*":
            return "rel_primitive_multiply"
        elif name == "=":
            return "rel_primitive_eq"
        elif name == "<=":
            return "rel_primitive_lt_eq"
        else:
            raise NotImplementedError(f"missing primitive case: {name}")

def _gen_rel_id(abstr: lqp.Abstraction) -> lqp.RelationId:
    return lqp.RelationId(hash_to_uint128(_lqp_hash(abstr)))

# TODO: this is NOT a good hash its just to get things working for now to get
# a stable id.
def _lqp_hash(node: lqp.LqpNode) -> int:
    if isinstance(node, lqp.Abstraction):
        h1 = _lqp_hash_list(node.vars)
        h2 = _lqp_hash(node.value)
        return _lqp_hash_fn((h1, h2))
    elif isinstance(node, lqp.Exists):
        h1 = _lqp_hash_list(node.vars)
        h2 = _lqp_hash(node.value)
        return _lqp_hash_fn((h1, h2))
    elif isinstance(node, lqp.Conjunction):
        h1 = _lqp_hash_list(node.args)
        return _lqp_hash_fn((h1,))
    elif isinstance(node, lqp.Var):
        return _lqp_hash_fn((node.name, node.type))
    elif isinstance(node, lqp.Constant):
        return _lqp_hash_fn((node.value,))
    elif isinstance(node, lqp.RelAtom):
        h1 = _lqp_hash(node.sig)
        h2 = _lqp_hash_list(node.terms)
        return _lqp_hash_fn((h1, h2))
    elif isinstance(node, lqp.RelationSig):
        return _lqp_hash_fn((node.name, tuple(node.types)))
    elif isinstance(node, lqp.Primitive):
        h1 = _lqp_hash_fn(node.name)
        h2 = _lqp_hash_list(node.terms)
        return _lqp_hash_fn((h1, h2))
    else:
        raise NotImplementedError(f"Unsupported LQP node type: {type(node)}")

# TODO: this is NOT a good hash its just to get things working for now to get
# a stable id.
def _lqp_hash_fn(node) -> int:
    return int.from_bytes(hashlib.sha256(str(node).encode()).digest(), byteorder='big', signed=False)

def _lqp_hash_list(node: Sequence[lqp.LqpNode]) -> int:
    hashes = [_lqp_hash(n) for n in node]
    return hash(tuple(hashes))

def hash_to_uint128(h: int) -> int:
    return h % (2**128)  # Ensure it's within the 128-bit range

# Preconditions
def _assert_valid_input(model: ir.Model) -> bool:
    # TODO: flesh this out more?
    _assert_root_is_logical(model.root)
    return True

def _assert_root_is_logical(task: ir.Task) -> bool:
    assert isinstance(task, ir.Logical), f"expected logical task, got {type(task)}"
    assert len(task.body) >= 1, f"expected at least one task, got {len(task.body)}"
    for subtask in task.body:
        # TODO: assert what subtasks should look like
        # TODO: where can aggregates exist? only on top level?
        assert isinstance(subtask, ir.Logical), f"expected logical task, got {type(subtask)}"

    return True

def type_from_var(arg: ir.Var) -> lqp.PrimitiveType:
    assert isinstance(arg, ir.Var)
    assert isinstance(arg.type, ir.ScalarType)
    if types.is_builtin(arg.type):
        # TODO: just ocompare to types.py
        if arg.type.name == "Int":
            return lqp.PrimitiveType.INT
        else:
            # TODO: fix this
            assert arg.type.name == "Any" or arg.type.name == "Number", f"Unknown type: {arg.type.name}"
            return lqp.PrimitiveType.UNKNOWN
    else:
        return lqp.PrimitiveType.UNKNOWN

def type_from_constant(arg: lqp.PrimitiveValue) -> lqp.PrimitiveType:
    if isinstance(arg, int):
        return lqp.PrimitiveType.INT
    elif isinstance(arg, float):
        return lqp.PrimitiveType.FLOAT
    elif isinstance(arg, str):
        return lqp.PrimitiveType.STRING
    else:
        raise NotImplementedError(f"Unknown constant type: {type(arg)}")



from typing import Optional

from relationalai.early_access.dsl.core.relations import Relation, AbstractRelation


class AbstractConcept:

    # We can add relation components to a ConceptModule by invoking it
    # with arguments that interleave reading text with the Types used
    # to play various Roles
    #
    def __setattr__(self, key, value):
        if key in dir(self) and key not in self.__dict__:
            raise Exception(f"Cannot override method {key} of Type {self._name} as an attribute.")
        else:
            if key[0] != '_':
                self._relations[key] = value
            return super().__setattr__(key, value)

    def __getitem__(self, key):
        return self.__dict__[key]

    def __init__(self, model, name):
        self._name = name
        self._model = model
        self._relations = {}

    def relation(self, *args, name: Optional[str] = None, functional: bool=False) -> Relation:
        if len(args) == 1 and not isinstance(args[0], str):
            raise ValueError("For binary or higher order relations parameter 'args' should contain "
                            "Sequence of text fragments followed by Types.")
        relationship = self._model.relationship(*[self, *args], relation_name=name)
        rel = next(iter(relationship.relations()), None)
        if rel is not None:
            rel.signature().set_functional(functional)
            return rel
        raise Exception(f"Could not find relation for relationship {relationship.name()}")

    def _add_relation(self, relation: AbstractRelation):
        self.__setattr__(relation._relname, relation)


from __future__ import annotations

from relationalai.early_access.metamodel import ir, factory as f, util, visitor
from relationalai.early_access.metamodel.compiler import Pass, group_tasks
from relationalai.early_access.metamodel.util import OrderedSet

class ExtractCommonLookups(Pass):
    """
    Extracts common lookups in the body of a logical as a new logical that derives those
    lookups in a temporary relation.

    From:
        Logical
            Logical
                lookup1
                lookup2
                Logical
                    x
                Logical
                    y
    To:
        Logical
            Logical
                lookup1
                lookup2
            Logical
                tmp
                Logical
                    x
                Logical
                    y
    """

    def rewrite(self, model: ir.Model, cache) -> ir.Model:
        if isinstance(model.root, ir.Logical):
            new_body = []
            new_relations:list[ir.Relation] = []
            for child in model.root.body:
                if isinstance(child, ir.Logical):
                    groups = group_tasks(child.body, {
                        "lookups": (ir.Lookup),
                        "logicals": (ir.Logical),
                        "aggregates": (ir.Aggregate),
                        "effects": (ir.Update, ir.Output)
                    })
                    # many lookups to extract because we have multiple logicals, and no aggregates
                    if len(groups["lookups"]) > 1 and len(groups["logicals"]) > 1 and not groups["aggregates"]:
                        new_logicals, relation = self._rewrite(child, groups)
                        new_body.extend(new_logicals)
                        new_relations.append(relation)
                    else:
                        new_body.append(child)
            if new_relations:
                new_relations.extend(model.relations)
                return ir.Model(
                    model.engines,
                    util.FrozenOrderedSet.from_iterable(new_relations),
                    model.types,
                    f.logical(new_body)
                )
        return model


    def _rewrite(self, node: ir.Logical, groups: dict[str, OrderedSet[ir.Task]]):
        new_logicals = []
        new_body = []

        # extract lookups as its own logical, deriving into a tmp relation
        extracted_rule, relation, exposed_vars = self._extract_common_lookups(node, groups)
        new_logicals.append(extracted_rule)

        lookup = ir.Lookup(node.engine, relation, tuple(exposed_vars.list))
        new_body.append(lookup)
        new_body.extend(groups["logicals"])
        if "aggregates" in groups:
            new_body.extend(groups["aggregates"])
        if "other" in groups:
            new_body.extend(groups["other"])
        if "effects" in groups:
            new_body.extend(groups["effects"])
        new_logicals.append(f.logical(new_body))

        return new_logicals, relation


    def _extract_common_lookups(self, node: ir.Logical, groups: dict[str, OrderedSet[ir.Task]]):
        lookups = groups["lookups"]
        effects = groups["effects"]

        # a bit of variable order optimization: if there's an effect in the node, try to use
        # the same variable order for the extracted relation
        lookup_vars = util.ordered_set()
        for lookup in lookups:
            lookup_vars.update(visitor.collect_vars(lookup))

        effect_vars = util.ordered_set()
        for e in effects:
            effect_vars.update(visitor.collect_vars(e))

        # first add the effect vars, and then the rest
        search_vars = util.ordered_set()
        for v in effect_vars:
            if v in lookup_vars:
                search_vars.add(v)
        search_vars.update(lookup_vars)

        # collect all variables used in the lookups, which are also used in other nodes
        exposed_vars: OrderedSet[ir.Var] = util.ordered_set()
        for v in search_vars:
            for task in node.body:
                if task not in lookups:
                    exposed_vars.add(v)

        # create a tmp relation to expose those vars
        relation = f.relation(f"common_{node.id}", [f.field(v.name, v.type) for v in exposed_vars])

        # create the new logical (a rule) that will derive into this relation
        # note that his logical reuses the lookup objects; exposed_vars also reuses the variable objects.
        body = []
        for lookup in lookups:
            body.append(lookup)
        body.append(f.derive(relation, exposed_vars.list))
        extracted_rule = f.logical(body)

        # return the extracted rule, the new temporary relation, and the variables exposed by it
        return extracted_rule, relation, exposed_vars



from __future__ import annotations
from dataclasses import dataclass
from enum import Enum
from typing import Union, Tuple, Sequence

# Tree representation of LQP. Each non-terminal (those with more than one
# option) is an "abstract" class and each terminal is its own class. All of
# which are children of LqpNode. PrimitiveType and PrimitiveValue are
# exceptions. PrimitiveType is an enum and PrimitiveValue is just a value.
# https://docs.google.com/document/d/1QXRU7zc1SUvYkyMCG0KZINZtFgzWsl9-XHxMssdXZzg/

@dataclass(frozen=True)
class LqpNode:
    # Return a string of this node in LLQP form.
    def to_llqp(self, indent_level=0):
        raise NotImplementedError(f"to_llqp not implemented for {type(self)}.")

# Declaration := Def | Loop
@dataclass(frozen=True)
class Declaration(LqpNode):
    pass

# TODO: this should probably be something else
@dataclass(frozen=True)
class LqpProgram(LqpNode):
    defs: list[Declaration]
    # name -> relation id
    outputs: list[Tuple[str, RelationId]]

    def to_llqp(self, indent_level=0):
        return Llqp.program_to_llqp(self, indent_level)

# Def(name::RelationId, body::Abstraction, attrs::Attribute[])
@dataclass(frozen=True)
class Def(Declaration):
    name: RelationId
    body: Abstraction
    attrs: Sequence[Attribute]

    def to_llqp(self, indent_level=0):
        return Llqp.def_to_llqp(self, indent_level)

# Loop(temporal_var::LoopIndex, inits::Def[], body::Declaration[])
@dataclass(frozen=True)
class Loop(Declaration):
    temporal_var: Var
    inits: Sequence[Def]
    body: Declaration

    def to_llqp(self, indent_level=0):
        return Llqp.loop_to_llqp(self, indent_level)

# Abstraction := Abstraction(vars::Var[], value::Formula)
@dataclass(frozen=True)
class Abstraction(LqpNode):
    vars: Sequence[Var]
    value: Formula

    def to_llqp(self, indent_level=0):
        return Llqp.abstraction_to_llqp(self, indent_level)

# Formula := Exists
#          | Reduce
#          | Conjunction
#          | Disjunction
#          | Not
#          | FFI
#          | Atom
#          | Pragma
#          | Primitive
#          | True
#          | False
#          | RelAtom
@dataclass(frozen=True)
class Formula(LqpNode):
    pass

# Exists(vars::Var[], value::Formula)
@dataclass(frozen=True)
class Exists(Formula):
    vars: Sequence[Var]
    value: Formula

    def to_llqp(self, indent_level=0):
        return Llqp.exists_to_llqp(self, indent_level)

# Reduce(op::Abstraction, body::Abstraction, terms::Term[])
@dataclass(frozen=True)
class Reduce(Formula):
    op: Abstraction
    body: Abstraction
    terms: Sequence[Term]

    def to_llqp(self, indent_level=0):
        return Llqp.reduce_to_llqp(self, indent_level)

# Conjunction(args::Formula[])
@dataclass(frozen=True)
class Conjunction(Formula):
    args: Sequence[Formula]

    def to_llqp(self, indent_level=0):
        return Llqp.conjunction_to_llqp(self, indent_level)

# Disjunction(args::Formula[])
@dataclass(frozen=True)
class Disjunction(Formula):
    args: Sequence[Formula]

    def to_llqp(self, indent_level=0):
        return Llqp.disjunction_to_llqp(self, indent_level)

# Not(arg::Formula)
@dataclass(frozen=True)
class Not(Formula):
    arg: Formula

    def to_llqp(self, indent_level=0):
        return Llqp.not_to_llqp(self, indent_level)

# FFI(name::Symbol, args::Abstraction[], terms::Term[])
@dataclass(frozen=True)
class Ffi(Formula):
    name: str
    args: Sequence[Abstraction]
    terms: Sequence[Term]

    def to_llqp(self, indent_level=0):
        return Llqp.ffi_to_llqp(self, indent_level)

# Atom(name::RelationId, terms::Term[])
@dataclass(frozen=True)
class Atom(Formula):
    name: RelationId
    terms: Sequence[Term]

    def to_llqp(self, indent_level=0):
        return Llqp.atom_to_llqp(self, indent_level)

# Pragma(name::Symbol, terms::Term[])
@dataclass(frozen=True)
class Pragma(Formula):
    name: str
    terms: Sequence[Term]

    def to_llqp(self, indent_level=0):
        return Llqp.pragma_to_llqp(self, indent_level)

# Primitive(name::Symbol, terms::Term[])
@dataclass(frozen=True)
class Primitive(Formula):
    name: str
    terms: Sequence[Term]

    def to_llqp(self, indent_level=0):
        return Llqp.primitive_to_llqp(self, indent_level)

# True()
@dataclass(frozen=True)
class JustTrue(Formula):
    def to_llqp(self, indent_level=0):
        return Llqp.true_to_llqp(self, indent_level)

# False()
@dataclass(frozen=True)
class JustFalse(Formula):
    def to_llqp(self, indent_level=0):
        return Llqp.false_to_llqp(self, indent_level)

# RelAtom(sig::RelationSig, terms::Term[])
@dataclass(frozen=True)
class RelAtom(Formula):
    sig: RelationSig
    terms: Sequence[Term]

    def to_llqp(self, indent_level=0):
        return Llqp.rel_atom_to_llqp(self, indent_level)

# Term := Var | Constant
@dataclass(frozen=True)
class Term(LqpNode):
    pass

# Var(name::Symbol, type::PrimitiveType)
@dataclass(frozen=True)
class Var(Term):
    name: str
    type: PrimitiveType

    def to_llqp(self, indent_level=0):
        return Llqp.var_to_llqp(self, indent_level)

# Constant(value::PrimitiveValue)
@dataclass(frozen=True)
class Constant(Term):
    value: PrimitiveValue

    def to_llqp(self, indent_level=0):
        return Llqp.constant_to_llqp(self, indent_level)

# Attribute := Attribute(name::Symbol, args::Constant[])
@dataclass(frozen=True)
class Attribute(LqpNode):
    name: str
    args: Sequence[Constant]

    def to_llqp(self, indent_level=0):
        return Llqp.attribute_to_llqp(self, indent_level)

# RelationId := RelationId(id::UInt128)
@dataclass(frozen=True)
class RelationId(LqpNode):
    # We use a catchall int here to represent the uint128 as it is difficult
    # to do so in Python without external packages. We check the value in
    # __post_init__.
    id: int

    def __post_init__(self):
        if self.id < 0 or self.id > 0xffffffffffffffffffffffffffffffff:
            raise ValueError(
                "RelationId constructed with out of range (UInt128) number: {}"
                    .format(self.id)
            )

    def to_llqp(self, indent_level=0):
        return Llqp.relation_id_to_llqp(self, indent_level)

# RelationSig := RelationSig(name::Symbol, types::PrimitiveType[])
@dataclass(frozen=True)
class RelationSig(LqpNode):
    name: str
    types: Sequence[PrimitiveType]

    def to_llqp(self, indent_level=0):
        return Llqp.relation_sig_to_llqp(self, indent_level)

# PrimitiveType := STRING | DECIMAL | INT | FLOAT | HASH
# TODO: we don't know what types we're supporting yet.
class PrimitiveType(Enum):
    # TODO: get rid of this oen maybe?
    UNKNOWN = 0
    STRING = 1
    INT = 2
    FLOAT = 3

    def to_llqp(self, indent_level=0):
        return Llqp.primitive_type_to_llqp(self, indent_level)

# PrimitiveValue := string | decimal | int | float | hash
# TODO: we don't know what types we're supporting yet.
PrimitiveValue = Union[str, int, float]

class Llqp:
    # Single INDentation.
    SIND = "    "

    # String of level indentations for LLQP.
    @staticmethod
    def indentation(level: int):
        return Llqp.SIND * level

    # Call .to_llqp on all nodes, each of which with indent_level, separating them
    # by delim.
    @staticmethod
    def list_to_llqp(nodes: Sequence[LqpNode], indent_level: int, delim: str):
        return delim.join(map(lambda n: n.to_llqp(indent_level), nodes))

    # Produces "(terms term1 term2 ...)" (all on one line) indented at indent_level.
    @staticmethod
    def terms_to_llqp(terms: Sequence[Term], indent_level: int):
        ind = Llqp.indentation(indent_level)

        llqp = ""
        if len(terms) == 0:
            llqp = ind + "(terms)"
        else:
            llqp = ind + "(terms " + Llqp.list_to_llqp(terms, 0, " ") + ")"

        return llqp

    ## .to_llqp implementations.

    @staticmethod
    def def_to_llqp(node: Def, indent_level: int):
        ind = Llqp.indentation(indent_level)

        llqp = ""
        llqp += ind + "(def " + node.name.to_llqp(0) + "\n"
        llqp += node.body.to_llqp(indent_level + 1) + "\n"
        if len(node.attrs) == 0:
            llqp += ind + Llqp.SIND + "(attrs) ;; no attrs" + "\n" + ind + ")"
        else:
            llqp += ind + Llqp.SIND + "(attrs" + "\n"
            llqp += Llqp.list_to_llqp(node.attrs, indent_level + 2, "\n") + "\n"
            llqp += ind + Llqp.SIND + ")" + ")"

        return llqp

    @staticmethod
    def loop_to_llqp(node: Loop, indent_level: int):
        ind = Llqp.indentation(indent_level)

        llqp = ""
        llqp += ind + "(loop" + node.temporal_var.to_llqp(0) + "\n"
        llqp += ind + Llqp.SIND + "(inits" + "\n"
        llqp += Llqp.list_to_llqp(node.inits, indent_level + 2, "\n") + "\n"
        llqp += ind + Llqp.SIND + ")" + "\n"
        llqp += ind + Llqp.SIND + "(body" + "\n"
        llqp += node.body.to_llqp(indent_level + 2) + "\n"
        llqp += ind + Llqp.SIND + ")" + ")"

        return llqp

    @staticmethod
    def abstraction_to_llqp(node: Abstraction, indent_level: int):
        ind = Llqp.indentation(indent_level)

        llqp = ""
        llqp += ind + "([" + Llqp.list_to_llqp(node.vars, 0, ", ") + "]" + "\n"
        llqp += node.value.to_llqp(indent_level + 1) + ")"

        return llqp

    @staticmethod
    def exists_to_llqp(node: Exists, indent_level: int):
        ind = Llqp.indentation(indent_level)

        llqp = ""
        llqp += ind + "(exists [" + Llqp.list_to_llqp(node.vars, 0, ", ") + "]" + "\n"
        llqp += node.value.to_llqp(indent_level + 1) + ")"

        return llqp

    @staticmethod
    def reduce_to_llqp(node: Reduce, indent_level: int):
        ind = Llqp.indentation(indent_level)

        llqp = ""
        llqp += ind + "(reduce" + " " + "\n"
        llqp += node.op.to_llqp(indent_level + 1) + "\n"
        llqp += node.body.to_llqp(indent_level + 1) + "\n"
        llqp += Llqp.terms_to_llqp(node.terms, indent_level + 1) + ")"

        return llqp

    @staticmethod
    def conjunction_to_llqp(node: Conjunction, indent_level: int):
        ind = Llqp.indentation(indent_level)

        llqp = ""
        llqp += ind + "(and" + "\n"
        llqp += Llqp.list_to_llqp(node.args, indent_level + 1, "\n") + ")"

        return llqp

    @staticmethod
    def disjunction_to_llqp(node: Disjunction, indent_level: int):
        ind = Llqp.indentation(indent_level)

        llqp = ""
        llqp += ind + "(or" + "\n"
        llqp += Llqp.list_to_llqp(node.args, indent_level + 1, "\n") + ")"

        return llqp

    @staticmethod
    def not_to_llqp(node: Not, indent_level: int):
        ind = Llqp.indentation(indent_level)

        llqp = ""
        llqp += ind + "(not" + "\n"
        llqp += node.arg.to_llqp(indent_level + 1) + ")"

        return llqp

    @staticmethod
    def ffi_to_llqp(node: Ffi, indent_level: int):
        ind = Llqp.indentation(indent_level)

        llqp = ""
        llqp += ind + "(ffi" + " " + ":" + node.name + "\n"
        llqp += ind + Llqp.SIND + "(args" + "\n"
        llqp += Llqp.list_to_llqp(node.args, indent_level + 2, "\n") + "\n"
        llqp += ind + Llqp.SIND + ")" + "\n"
        llqp += Llqp.terms_to_llqp(node.terms, indent_level + 1) + ")"

    @staticmethod
    def atom_to_llqp(node: Atom, indent_level: int):
        ind = Llqp.indentation(indent_level)
        return f"{ind}(atom {node.name.to_llqp(0)} {Llqp.list_to_llqp(node.terms, 0, ' ')})"

    @staticmethod
    def pragma_to_llqp(node: Pragma, indent_level: int):
        ind = Llqp.indentation(indent_level)
        return f"{ind}(pragma :{node.name} {Llqp.terms_to_llqp(node.terms, 0)})"

    @staticmethod
    def primitive_to_llqp(node: Primitive, indent_level: int):
        ind = Llqp.indentation(indent_level)
        return f"{ind}(primitive :{node.name} {Llqp.list_to_llqp(node.terms, 0, ' ')})"

    @staticmethod
    def true_to_llqp(node: JustTrue, indent_level: int):
        ind = Llqp.indentation(indent_level)
        return ind + "(true)"

    @staticmethod
    def false_to_llqp(node: JustFalse, indent_level: int):
        ind = Llqp.indentation(indent_level)
        return ind + "(false)"

    @staticmethod
    def rel_atom_to_llqp(node: RelAtom, indent_level: int):
        ind = Llqp.indentation(indent_level)
        return f"{ind}(relatom {node.sig.to_llqp(0)} {Llqp.list_to_llqp(node.terms, 0, ' ')})"

    @staticmethod
    def var_to_llqp(node: Var, indent_level: int):
        ind = Llqp.indentation(indent_level)
        return ind + node.name + "::" + node.type.to_llqp(0)

    @staticmethod
    def constant_to_llqp(node: Constant, indent_level: int):
        ind = Llqp.indentation(indent_level)

        llqp = ""
        llqp += ind
        if isinstance(node.value, str):
            llqp += "\"" + node.value + "\""
        else:
            # suffices to just dump the value?
            llqp += str(node.value)

        return llqp

    @staticmethod
    def attribute_to_llqp(node: Attribute, indent_level: int):
        ind = Llqp.indentation(indent_level)

        llqp = ""
        llqp += ind
        llqp += "(attribute" + " "
        llqp += ":" + node.name + " "
        if len(node.args) == 0:
            llqp += "(args)"
        else:
            llqp += "(args" + " "
            llqp += Llqp.list_to_llqp(node.args, 0, " ")
            llqp += ")"
        llqp += ")"

        return llqp

    @staticmethod
    def relation_id_to_llqp(node: RelationId, indent_level: int):
        ind = Llqp.indentation(indent_level)
        return f"{ind}:{str(node.id)}"

    @staticmethod
    def relation_sig_to_llqp(node: RelationSig, indent_level: int):
        ind = Llqp.indentation(indent_level)
        llqp = ""
        llqp += ind
        llqp += "(sig" + " "
        llqp += ":" + node.name
        for type in node.types:
            llqp += "/" + type.to_llqp(0)
        llqp += ")"

        return llqp

    @staticmethod
    def primitive_type_to_llqp(node: PrimitiveType, indent_level: int):
        ind = Llqp.indentation(indent_level)
        return ind + node.name

    @staticmethod
    def program_to_llqp(node: LqpProgram, indent_level: int):
        # TODO: is this true? and in general for the other things can they be missing?
        assert len(node.defs) > 0, "LqpProgram must have at least one def."
        reads_portion = ""
        if len(node.outputs) == 0:
            reads_portion += _raw_text("(reads) ;; no outputs", indent_level + 2) + "\n"
        else:
            reads_portion += _raw_text("(reads\n", indent_level + 2)

            for (name, rel_id) in node.outputs:
                reads_portion +=\
                    f"{Llqp.indentation(indent_level + 3)}" +\
                    f"({name} " +\
                    f"{Llqp.relation_id_to_llqp(rel_id, 0)}" +\
                    ")\n"

            reads_portion += _raw_text(")\n", indent_level + 2)

        delim = "\n\n"
        writes_portion = f"{Llqp.list_to_llqp(node.defs, indent_level + 5, delim)}\n"


        return\
        _raw_text("(transaction\n", indent_level) +\
        _raw_text("(epoch\n", indent_level + 1) +\
        _raw_text("(local_writes\n", indent_level + 2) +\
        _raw_text("(define\n", indent_level + 3) +\
        _raw_text("(fragment :f1\n", indent_level + 4) +\
        writes_portion +\
        _raw_text(")\n", indent_level + 4) +\
        _raw_text(")\n", indent_level + 3) +\
        _raw_text(")\n", indent_level + 2) +\
        "\n" +\
        reads_portion +\
        _raw_text(")\n", indent_level + 1) +\
        _raw_text(")\n", indent_level)

def _raw_text(txt: str, indent_level: int):
        ind = Llqp.indentation(indent_level)
        return f"{ind}{txt}"



from __future__ import annotations
import functools
from typing import Any, Callable

from relationalai.errors import AsBoolForNonFilter
from relationalai.metamodel import Builtins
from ..dsl import alias, rel, Vars, create_vars, create_var, tag
from .. import dsl, metamodel as m
from . import graphs, aggregates, strings, math, dates

#--------------------------------------------------
# Tagging some base rel functions
#--------------------------------------------------

tag(rel.decimal, Builtins.SingleValued)
tag(rel.float, Builtins.SingleValued)
tag(rel.int, Builtins.SingleValued)
tag(rel.sqrt, Builtins.SingleValued)
tag(rel.concat, Builtins.SingleValued)

#--------------------------------------------------
# Utilities
#--------------------------------------------------

def as_bool(expr: dsl.Expression) -> dsl.Expression:
    if expr._expr.entity.isa(Builtins.Filter):
        # add the filter to apply
        prev_filter = expr._expr.entity.value
        filter = dsl.build.property_named("filter", [])
        expr._expr.append(filter, m.Var(value=prev_filter))
        # add a result var
        expr._var = m.Var(Builtins.Unknown)
        prop = dsl.build.property_named("result", [])
        expr._expr.append(prop, expr._var)
        # use pyrel_bool_filter to apply the filter and get a bool
        expr._expr.entity = m.Var(value=Builtins.BoolFilter)
    else:
        raise AsBoolForNonFilter()
    return expr

def as_rows(data:list[tuple|dict|int|float|str]) -> dsl.Rows:
    return dsl.Rows(dsl.get_graph(), data)

def _wrapped(relation: Callable, *args: Any) -> Callable:
    def f(*args: Any):
        res = create_var()
        relation(*args, res)
        return res
    return f

def minimum(arg1: Any, arg2: Any, *args: Any) -> dsl.Expression:
    return functools.reduce(_wrapped(rel.minimum), [arg1, arg2, *args])

def maximum(arg1: Any, arg2: Any, *args: Any) -> dsl.Expression:
    return functools.reduce(_wrapped(rel.maximum), [arg1, arg2, *args])

__all__ = [
    "aggregates",
    "alias",
    "dates",
    "graphs",
    "math",
    "minimum",
    "maximum",
    "rel",
    "strings",
    "Vars",
    "create_vars",
    "create_var"
]
